
import sys
import unittest
import array
import pickle
import operator
from test import test_support

sys.path.append("..")

import creator
import base
import gp
import tools

def func():
    return "True"

class Pickling(unittest.TestCase):

    def setUp(self):
        creator.create("FitnessMax", base.Fitness, weights=(1.0,))
        creator.create("IndList", list, fitness=creator.FitnessMax)
        creator.create("IndArray", array.array,  typecode='f', fitness=creator.FitnessMax)
        creator.create("IndTree", base.Tree, fitness=creator.FitnessMax)
        self.toolbox = base.Toolbox()
        self.toolbox.register("func", func)
        self.toolbox.register("lambda_func", lambda: "True")
    
    def test_pickle_fitness(self):
        fitness = creator.FitnessMax()
        fitness.values = (1.0,)
        fitness_s = pickle.dumps(fitness)
        fitness_l = pickle.loads(fitness_s)
        self.failUnlessEqual(fitness, fitness_l, "Unpickled fitness != pickled fitness")

    def test_pickle_ind_list(self):
        ind = creator.IndList([1.0, 2.0, 3.0])
        ind.fitness.values = (4.0,)
        ind_s = pickle.dumps(ind)
        ind_l = pickle.loads(ind_s)
        self.failUnlessEqual(ind, ind_l, "Unpickled individual list != pickled individual list")
        self.failUnlessEqual(ind.fitness, ind_l.fitness, "Unpickled individual fitness != pickled individual fitness")
         
    def test_pickle_ind_array(self):
        ind = creator.IndArray([1.0, 2.0, 3.0])
        ind.fitness.values = (4.0,)
        ind_s = pickle.dumps(ind)
        ind_l = pickle.loads(ind_s)
        self.failUnlessEqual(ind, ind_l, "Unpickled individual array != pickled individual array")
        self.failUnlessEqual(ind.fitness, ind_l.fitness, "Unpickled individual fitness != pickled individual fitness")
    
    def test_pickle_tree(self):
        ind = creator.IndTree(["+", 1, 2])
        ind.fitness.values = (1.0,)
        ind_s = pickle.dumps(ind)
        ind_l = pickle.loads(ind_s)
        msg =  "Unpickled individual %s != pickled individual %s" % (str(ind), str(ind_l))
        self.failUnlessEqual(ind, ind_l, msg)
        msg =  "Unpickled fitness %s != pickled fitness %s" % (str(ind.fitness), str(ind_l.fitness))
        self.failUnlessEqual(ind.fitness, ind_l.fitness, msg)
    
    def test_pickle_population(self):
        ind1 = creator.IndList([1.0,2.0,3.0])
        ind1.fitness.values = (1.0,)
        ind2 = creator.IndList([4.0,5.0,6.0])
        ind2.fitness.values = (2.0,)
        ind3 = creator.IndList([7.0,8.0,9.0])
        ind3.fitness.values = (3.0,)
        
        pop = [ind1, ind2, ind3]
        
        pop_s = pickle.dumps(pop)
        pop_l = pickle.loads(pop_s)
        
        self.failUnlessEqual(pop[0], pop_l[0], "Unpickled individual list != pickled individual list")
        self.failUnlessEqual(pop[0].fitness, pop_l[0].fitness, "Unpickled individual fitness != pickled individual fitness")
        self.failUnlessEqual(pop[1], pop_l[1], "Unpickled individual list != pickled individual list")
        self.failUnlessEqual(pop[1].fitness, pop_l[1].fitness, "Unpickled individual fitness != pickled individual fitness")
        self.failUnlessEqual(pop[2], pop_l[2], "Unpickled individual list != pickled individual list")
        self.failUnlessEqual(pop[2].fitness, pop_l[2].fitness, "Unpickled individual fitness != pickled individual fitness")
    
    if not sys.version_info < (2, 7):
        def test_pickle_partial(self):
            func_s = pickle.dumps(self.toolbox.func)
            func_l = pickle.loads(func_s)

            self.failUnlessEqual(self.toolbox.func(), func_l())
        
        @unittest.expectedFailure
        def test_pickle_lambda(self):
            func_s = pickle.dumps(self.toolbox.lambda_func)
            func_l = pickle.loads(func_s)
        
            self.failUnlessEqual(self.toolbox.lambda_func(), func_l())

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(Pickling)
    unittest.TextTestRunner(verbosity=2).run(suite)



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

"""
A very simple example of recursive nested tasks.
Each task maps 2 others tasks, each of these 2 tasks maps 2 others, etc.,
up to RECURSIVITY_LEVEL.

Setting RECURSIVITY_LEVEL to a high value (typically more than 16) may
produce some strange OS errors (as the kernel will probably complain about
the number of active threads...
"""

from deap import dtm

RECURSIVITY_LEVEL = 12

def recursiveFunc(level):
    if level == 1:
        return 1
    else:
        args = [level-1] * 2
        s = sum(dtm.map(recursiveFunc, args))
        if level == RECURSIVITY_LEVEL:
            print("2^"+str(level)+" = " + str(s))
        return s

dtm.start(recursiveFunc, RECURSIVITY_LEVEL)



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import sys
import random
import logging

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

from deap import algorithms
from deap import base
from deap import creator
from deap import tools

import sortingnetwork as sn

INPUTS = 6

def evalEvoSN(individual, dimension):
    network = sn.SortingNetwork(dimension, individual)
    return network.assess(), network.length, network.depth

def genWire(dimension):
    return (random.randrange(dimension), random.randrange(dimension))
    
def genNetwork(dimension, min_size, max_size):
    size = random.randint(min_size, max_size)
    return [genWire(dimension) for i in xrange(size)]
    
def mutWire(individual, dimension, indpb):
    for index, elem in enumerate(individual):
        if random.random() < indpb:
            individual[index] = genWire(dimension)      

def mutAddWire(individual, dimension):
    index = random.randint(0, len(individual))
    individual.insert(index, genWire(dimension))

def mutDelWire(individual):
    index = random.randrange(len(individual))
    del individual[index]

creator.create("FitnessMin", base.Fitness, weights=(-1.0, -1.0, -1.0))
creator.create("Individual", list, fitness=creator.FitnessMin)

toolbox = base.Toolbox()

# Gene initializer
toolbox.register("network", genNetwork, dimension=INPUTS, min_size=9, max_size=12)

# Structure initializers
toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.network)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

toolbox.register("evaluate", evalEvoSN, dimension=INPUTS)
toolbox.register("mate", tools.cxTwoPoints)
toolbox.register("mutate", mutWire, dimension=INPUTS, indpb=0.05)
toolbox.register("addwire", mutAddWire, dimension=INPUTS)
toolbox.register("delwire", mutDelWire)
toolbox.register("select", tools.selNSGA2)

def main():
    random.seed(64)

    population = toolbox.population(n=300)
    hof = tools.ParetoFront()
    
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)

    CXPB, MUTPB, ADDPB, DELPB, NGEN = 0.5, 0.2, 0.01, 0.01, 40
    
    # Evaluate every individuals
    fitnesses = toolbox.map(toolbox.evaluate, population)
    for ind, fit in zip(population, fitnesses):
        ind.fitness.values = fit
    
    hof.update(population)
    stats.update(population)
    
    # Begin the evolution
    for g in xrange(NGEN):
        print "-- Generation %i --" % g
        offsprings = [toolbox.clone(ind) for ind in population]
    
        # Apply crossover and mutation
        for ind1, ind2 in zip(offsprings[::2], offsprings[1::2]):
            if random.random() < CXPB:
                toolbox.mate(ind1, ind2)
                del ind1.fitness.values
                del ind2.fitness.values
        
        # Note here that we have a different sheme of mutation than in the
        # original algorithm, we use 3 different mutations subsequently.
        for ind in offsprings:
            if random.random() < MUTPB:
                toolbox.mutate(ind)
                del ind.fitness.values
            if random.random() < ADDPB:
                toolbox.addwire(ind)
                del ind.fitness.values
            if random.random() < DELPB:
                toolbox.delwire(ind)
                del ind.fitness.values
                
        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offsprings if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit
        
        print "  Evaluated %i individuals" % len(invalid_ind)
        
        population = toolbox.select(population+offsprings, len(offsprings))
        hof.update(population)
        stats.update(population)
        print stats

    best_network = sn.SortingNetwork(INPUTS, hof[0])
    print best_network
    print best_network.draw()
    print "%i errors, length %i, depth %i" % hof[0].fitness.values
    
    return population, stats, hof

if __name__ == "__main__":
    main()



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

"""The :mod:`~deap.creator` module is the heart and soul of DEAP, it allows to
create, at runtime, classes that will fulfill the needs of your evolutionary
algorithms. This module follows the meta-factory paradigm by allowing to
create new classes via both composition and inheritance. Attributes both datas
and functions are added to existing types in order to create new types
empowered with user specific evolutionary computation capabilities. In effect,
new classes can be built from any imaginable type, from :class:`list` to
:class:`set`, :class:`dict`, :class:`~deap.base.Tree` and more,
providing the possibility to implement genetic algorithms, genetic
programming, evolution strategies, particle swarm optimizers, and many more.
"""

import array
import copy

class_replacers = {}
"""Some classes in Python's standard library as well as third party library
may be in part incompatible with the logic used in DEAP. In order to palliate
to this problem, the method :func:`create` uses the dictionary
`class_replacers` to identify if the base type provided is problematic, and if
so  the new class inherits from the replacement class instead of the
original base class.

`class_replacers` keys are classes to be replaced and the values are the
replacing classes.
"""

try:
    import numpy
    (numpy.ndarray, numpy.array)
except ImportError:
    # Numpy is not present, skip the definition of the replacement class.
    pass
except AttributeError:
    # Numpy is present, but there is either no ndarray or array in numpy,
    # also skip the definition of the replacement class.
    pass
else:
    class _numpy_array(numpy.ndarray):
        def __getslice__(self, i, j):
            """Overrides the getslice from numpy.ndarray that returns a shallow
            copy of the slice.
            """
            return numpy.ndarray.__getslice__(self, i, j).copy()
        
        def __deepcopy__(self, memo):
            """Overrides the deepcopy from numpy.ndarray that does not copy
            the object's attributes.
            """
            copy_ = numpy.ndarray.__deepcopy__(self, memo)
            copy_.__dict__.update(copy.deepcopy(self.__dict__, memo))
            return copy_

        @staticmethod
        def __new__(cls, iterable):
            """Creates a new instance of a numpy.ndarray from a function call"""
            return numpy.array(list(iterable)).view(cls)

        def __array_finalize__(self, obj):
            # __init__ will reinitialize every member of the subclass.
            # this might not be desirable for example in the case of an ES. 
            self.__init__()

            # Instead, e could use the following that will simply deepcopy 
            # every member that is present in the original class
            # This is significantly slower. 
            #if self.__class__ == obj.__class__:
            #    self.__dict__.update(copy.deepcopy(obj.__dict__))
    class_replacers[numpy.ndarray] = _numpy_array

class _array(array.array):
    @staticmethod
    def __new__(cls, seq=()):
        return super(_array, cls).__new__(cls, cls.typecode, seq)
    
    def __deepcopy__(self, memo):
        """Overrides the deepcopy from array.array that does not copy
        the object's attributes and class type.
        """
        cls = self.__class__
        copy_ = cls.__new__(cls, self)
        memo[id(self)] = copy_
        copy_.__dict__.update(copy.deepcopy(self.__dict__, memo))
        return copy_

    def __reduce__(self):
        return (self.__class__, (list(self),), self.__dict__)
class_replacers[array.array] = _array

def create(name, base, **kargs):
    """Creates a new class named *name* inheriting from *base* in the
    :mod:`~deap.creator` module. The new class can have attributes defined by
    the subsequent keyword arguments passed to the function create. If the
    argument is a class (without the parenthesis), the __init__ function is
    called in the initialization of an instance of the new object and the
    returned instance is added as an attribute of the class' instance.
    Otherwise, if the argument is not a class, (for example an :class:`int`),
    it is added as a "static" attribute of the class.
    
    The following is used to create a class :class:`Foo` inheriting from the
    standard :class:`list` and having an attribute :attr:`bar` being an empty
    dictionary and a static attribute :attr:`spam` initialized to 1. ::
    
        create("Foo", list, bar=dict, spam=1)
        
    This above line is exactly the same as defining in the :mod:`creator`
    module something like the following. ::
    
        def Foo(list):
            spam = 1
            
            def __init__(self):
                self.bar = dict()
    """
    dict_inst = {}
    dict_cls = {}
    for obj_name, obj in kargs.iteritems():
        if hasattr(obj, "__call__"):
            dict_inst[obj_name] = obj
        else:
            dict_cls[obj_name] = obj

    # Check if the base class has to be replaced
    if base in class_replacers:
        base = class_replacers[base]

    # A DeprecationWarning is raised when the object inherits from the 
    # class "object" which leave the option of passing arguments, but
    # raise a warning stating that it will eventually stop permitting
    # this option. Usually this happens when the base class does not
    # override the __init__ method from object.
    def initType(self, *args, **kargs):
        """Replace the __init__ function of the new type, in order to
        add attributes that were defined with **kargs to the instance.
        """
        for obj_name, obj in dict_inst.iteritems():
            setattr(self, obj_name, obj())
        if base.__init__ is not object.__init__:
            base.__init__(self, *args, **kargs)
        else:
            base.__init__(self)

    objtype = type(name, (base,), dict_cls)
    objtype.__init__ = initType
    globals()[name] = objtype



#    This file is part of EAP.
#
#    EAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    EAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with EAP. If not, see <http://www.gnu.org/licenses/>.

import sys
import random
import operator
import logging

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap import gp

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

def if_then_else(condition, out1, out2):
    return out1 if condition else out2

# Initialize Multiplexer problem input and output vectors

MUX_SELECT_LINES = 3
MUX_IN_LINES = 2 ** MUX_SELECT_LINES
MUX_TOTAL_LINES = MUX_SELECT_LINES + MUX_IN_LINES

# input : [A0 A1 A2 D0 D1 D2 D3 D4 D5 D6 D7] for a 8-3 mux
inputs = [[0] * MUX_TOTAL_LINES for i in range(2 ** MUX_TOTAL_LINES)]
outputs = [None] * (2 ** MUX_TOTAL_LINES)

for i in range(2 ** MUX_TOTAL_LINES):
    value = i
    divisor = 2 ** MUX_TOTAL_LINES
    # Fill the input bits
    for j in range(MUX_TOTAL_LINES):
        divisor /= 2
        if value >= divisor:
            inputs[i][j] = 1
            value -= divisor
    
    # Determine the corresponding output
    indexOutput = MUX_SELECT_LINES
    for j, k in enumerate(inputs[i][:MUX_SELECT_LINES]):
        if k:   indexOutput += 2 ** j
    outputs[i] = inputs[i][indexOutput]

pset = gp.PrimitiveSet("MAIN", MUX_TOTAL_LINES, "IN")
pset.addPrimitive(operator.and_, 2)
pset.addPrimitive(operator.or_, 2)
pset.addPrimitive(operator.not_, 1)
pset.addPrimitive(if_then_else, 3)
pset.addTerminal(1)
pset.addTerminal(0)

creator.create("FitnessMax", base.Fitness, weights=(1.0,))
creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMax, pset=pset)

toolbox = base.Toolbox()
toolbox.register("expr", gp.genFull, pset=pset, min_=2, max_=4)
toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)
toolbox.register("lambdify", gp.lambdify, pset=pset)

def evalMultiplexer(individual):
    func = toolbox.lambdify(expr=individual)
    good = sum((func(*(inputs[i])) == outputs[i] for i in range(2 ** MUX_TOTAL_LINES)))
    return good,

toolbox.register("evaluate", evalMultiplexer)
toolbox.register("select", tools.selTournament, tournsize=7)
toolbox.register("mate", gp.cxUniformOnePoint)
toolbox.register("expr_mut", gp.genGrow, min_=0, max_=2)
toolbox.register("mutate", gp.mutUniform, expr=toolbox.expr_mut)

def main():
    random.seed()
    pop = toolbox.population(n=500)
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)
    
    algorithms.eaSimple(toolbox, pop, 0.8, 0.1, 40, stats, halloffame=hof)
    
    logging.info("Best individual is %s, %s", gp.evaluate(hof[0]), hof[0].fitness)
    
    return pop, stats, hof

if __name__ == "__main__":
    main()




import threading

from abc import ABCMeta, abstractmethod, abstractproperty

class AbstractCommThread(threading.Thread):
    __metaclass__ = ABCMeta

    def __init__(self, recvQ, sendQ, mainThreadEvent, exitEvent, commReadyEvent, randomGenerator, cmdlineArgs):
        threading.Thread.__init__(self)
        self.recvQ = recvQ
        self.sendQ = sendQ
        
        self.exitStatus = exitEvent
        self.msgSendTag = 2
        self.wakeUpMainThread = mainThreadEvent
        self.random = randomGenerator
        self.commReadyEvent = commReadyEvent
        
        self.cmdArgs = cmdlineArgs
        
        self.traceMode = False
        self.traceTo = None
    
    @abstractproperty
    def poolSize(self):
        """
        Return the number of effective workers (for instance, with MPI, this
        is the number of slots asked with the -n or -np option)
        """
        pass

    @abstractproperty
    def workerId(self):
        """
        Return an ID for this worker such as each worker gets a different ID.
        This must be an immutable type (int, string, tuple, etc.)
        """
        pass

    @abstractproperty
    def isRootWorker(self):
        """
        Return True if this worker is the "root worker", that is the worker
        which will start the main task. The position of this worker in the
        hosts is not important, but one and only one worker should be
        designated as the root worker (the others should receive False).
        """
        pass
    
    @abstractproperty
    def isLaunchProcess(self):
        """
        If this function returns True, the main thread will wait for the
        termination of this thread, and then exit without executing any
        task. This may be useful for the backends which have to launch
        themselves the remote DTM processes.
        """
        pass
    
    @abstractmethod
    def setTraceModeOn(self, xmlLogger):
        """
        Used for logging purposes. The xmlLogger arg is an xml.etree object
        which can be use by the backend to log some informations. The log
        format is not currently specified, and the backend may choose to
        ignore this call (and not log anything).
        """
        pass
    
    @abstractmethod
    def iterOverIDs(self):
        """
        Return an iterable over all the worker IDs. For instance, if your
        workers IDs are integers between 0 and 63, it should then return
        range(0,64). 
        """
        pass
    
    @abstractmethod
    def run(self):
        """
        The main method, which will be call to start the communication backend.
        This is the place where you should import your special modules, and
        insert the communication loop.
        """
        pass
        


#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import threading
import time
import math
import random
try:
    import Queue
except ImportError:
    import queue as Queue
import copy
import pickle
import logging
import sys
import os

PRETTY_PRINT_SUPPORT = False
try:
    from lxml import etree
    PRETTY_PRINT_SUPPORT = True
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        # Python 2.5
        import xml.etree.ElementTree as etree

from deap.dtm.dtmTypes import *

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
_logger = logging.getLogger("dtm.control")

# Constantes
DTM_CONTROL_THREAD_LATENCY = 0.05

MSG_COMM_TYPE = 0
MSG_SENDER_INFO = 1
MSG_NODES_INFOS = 2

DTM_LOGDIR_DEFAULT_NAME = "DtmLog"
            
try:
    from math import erf
except ImportError:
    def erf(x):
        # See http://stackoverflow.com/questions/457408/is-there-an-easily-available-implementation-of-erf-for-python
        # save the sign of x
        sign = 1
        if x < 0:
            sign = -1
        x = abs(x)

        # constants
        a1 = 0.254829592
        a2 = -0.284496736
        a3 = 1.421413741
        a4 = -1.453152027
        a5 = 1.061405429
        p = 0.3275911

        # A&S formula 7.1.26
        t = 1.0 / (1.0 + p * x)
        y = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * math.exp(-x * x)
        return sign * y # erf(-x) = -erf(x)


class TaskIdGenerator(object):
    """
    A thread-safe ID generator.
    A DTM task ID is a tuple looking like (workerId, uniqId)
    With MPI, workerId is an integer (rank) and the uniqIds start from 0,
    but workerId can be virtually any immutable object (as it is used as
    a dictionnary key)
    """
    def __init__(self, rank, initId=0):
        self.r = rank
        self.wtid = 0
        self.generatorLock = threading.Lock()

    @property
    def tid(self):
        self.generatorLock.acquire()
        newId = self.wtid
        self.wtid += 1
        self.generatorLock.release()
        return (self.r, newId)


class ExecInfo(object):
    """
    Contains the information about the current task
    """
    def __init__(self, tStats, tStatsLock):
        self.eLock = threading.Lock()
        self.tStats = tStats
        self.tStatsLock = tStatsLock
        self.eCurrent = None
        self.mainThread = threading.currentThread()
        self.piSqrt = math.sqrt(math.pi)
        self.sqrt2 = 2 ** 0.5

    def _execTimeRemaining(self, mean, stdDev, alreadyDone):
        if alreadyDone == 0.:
            # If not started yet, return mean
            return mean
        #
        # Else compute the mass center of the remaining part of the gaussian
        # For instance, if we have a gaussian (mu, sigma) = (3, 4)
        # and that the task is executing for 3 seconds, then we may estimate
        # that it will probably finished at :
        # int(x*gaussian) / int(gaussian) over [3, +inf[
        # (where int is integral)
        #
        # We get 6.192 seconds, that is an expected remaining time of
        # 6.192 - 3 = 3.192 seconds
        
        # Evaluate primitive at point 'alreadyDone'
        commonPart = erf(self.sqrt2 * (alreadyDone - mean) / (stdDev * 2))
        areaPart1 = 0.5 * commonPart
        massCenterPart1 = (self.sqrt2 / (4 * self.piSqrt)) * (-2 * stdDev * math.exp(-0.5 * ((alreadyDone - mean) ** 2) / (stdDev ** 2)) + mean * (self.piSqrt) * (self.sqrt2) * commonPart)

        # Evaluate primitive at the infinite
        # erf(+inf) = 1, so
        areaPart2 = 0.5
        # exp(-inf) = 0, and erf(+inf) = 1, so
        massCenterPart2 = mean / 2.

        if areaPart1 == areaPart2:
            # So far from the mean that erf(alreadyDone - moyenne) = 1
            previsionPoint = alreadyDone + 0.5      # Hack : lets assume that there's 0.5 sec left...
        else:
            previsionPoint = (massCenterPart2 - massCenterPart1) / (areaPart2 - areaPart1)

        return previsionPoint - alreadyDone


    def acquire(self, forTask, blocking=True):
        if self.eLock.acquire(blocking):
            self.eCurrent = forTask
            return True
        else:
            return False

    def isLocked(self):
        if self.eLock.acquire(False):
            self.eLock.release()
            return False
        return True

    def release(self):
        self.eCurrent = None
        self.eLock.release()

    def getLoad(self):
        try:
            self.tStatsLock.acquire()
            tInfo = self.tStats.get(self.eCurrent.target, StatsContainer(rAvg=1., rStdDev=1., rSquareSum=0., execCount=0))
            val = self._execTimeRemaining(tInfo.rAvg, tInfo.rStdDev, self.eCurrent.threadObject.timeExec)
            self.tStatsLock.release()
            return val
        except AttributeError:
            self.tStatsLock.release()
            return 0.


class TaskQueue(object):
    """
    """
    def __init__(self, tasksStatsStruct, tasksStatsLock):
        self.tStats = tasksStatsStruct
        self.tStatsLock = tasksStatsLock
        self.previousLoad = 0.
        self.piSqrt = math.sqrt(math.pi)
        self.sqrt2 = 2 ** 0.5
        self._tQueue = Queue.PriorityQueue()
        self._tDict = {}
        self._actionLock = threading.Lock()
        self.changed = True

    def _getTimeInfoAbout(self, task):
        if task.threadObject is None:
            timeDone = 0.
        else:
            timeDone = task.threadObject.timeExec

        self.tStatsLock.acquire()
        tInfo = self.tStats.get(task.target, StatsContainer(rAvg=1., rStdDev=1., rSquareSum=0., execCount=0))
        self.tStatsLock.release()

        return self._execTimeRemaining(tInfo.rAvg, tInfo.rStdDev, timeDone)

    def _execTimeRemaining(self, mean, stdDev, alreadyDone):
        if alreadyDone == 0. or stdDev == 0:
            return mean

        #
        # Else compute the mass center of the remaining part of the gaussian
        # For instance, if we have a gaussian (mu, sigma) = (3, 4)
        # and that the task is executing for 3 seconds, then we may estimate
        # that it will probably finished at :
        # int(x*gaussian) / int(gaussian) over [3, +inf[
        # (where int is integral)
        #
        # We get 6.192 seconds, that is an expected remaining time of
        # 6.192 - 3 = 3.192 seconds

        # Evaluate primitive at point 'alreadyDone'
        commonPart = erf(self.sqrt2 * (alreadyDone - mean) / (stdDev * 2))
        areaPart1 = 0.5 * commonPart
        massCenterPart1 = (self.sqrt2 / (4 * self.piSqrt)) * (-2 * stdDev * math.exp(-0.5 * ((alreadyDone - mean) ** 2) / (stdDev ** 2)) + mean * (self.piSqrt) * (self.sqrt2) * commonPart)

        # Evaluate primitive at the infinite
        # erf(+inf) = 1, so
        areaPart2 = 0.5
        # exp(-inf) = 0, and erf(+inf) = 1, so
        massCenterPart2 = mean / 2.

        if areaPart1 == areaPart2:
            # So far from the mean that erf(alreadyDone - moyenne) = 1
            previsionPoint = alreadyDone + 0.5      # Hack : lets assume that there's 0.5 sec left...
        else:
            previsionPoint = (massCenterPart2 - massCenterPart1) / (areaPart2 - areaPart1)

        return previsionPoint - alreadyDone

    def put(self, taskObject):
        self.putList((taskObject,))

    def putList(self, tasksList):
        self._actionLock.acquire()

        for taskObject in tasksList:
            self._tDict[taskObject.tid] = taskObject
            self._tQueue.put(taskObject)
        
        self.changed = True
        self._actionLock.release()

    def getTask(self):
        self._actionLock.acquire()
        while True:
            try:
                taskObject = self._tQueue.get_nowait()
            except Queue.Empty:
                self._actionLock.release()
                raise Queue.Empty

            if taskObject.tid in self._tDict:
                del self._tDict[taskObject.tid]
                self.changed = True
                self._actionLock.release()
                return taskObject

    def isTaskIn(self, taskId):
        return taskId in self._tDict

    def _getApproximateLoad(self):
        return self.previousLoad
    
    def getLen(self):
        return len(self._tDict)
    
    def getLoad(self):
        tmpLoad = 0.
        self._actionLock.acquire()
        if not self.changed:
            self._actionLock.release()
            return self.previousLoad
	  
        for tid, tObj in self._tDict.items():
            tmpLoad += self._getTimeInfoAbout(tObj)

        self.previousLoad = tmpLoad
        self.changed = False
        self._actionLock.release()
        return self.previousLoad

    def getSpecificTask(self, taskId):
        self._actionLock.acquire()
        if taskId in self._tDict:
            task = self._tDict[taskId]
            del self._tDict[taskId]
            self.changed = True
        else:
            task = None
        self._actionLock.release()
        return task

    def getTaskByExecTime(self, execTimeWanted, maxDiff= -1.):
        # maxDiff is the max difference wanted
        # If there is no task with the execTimeWanted +- maxDiff,
        # return None
        # maxDiff = -1 means that there is no max difference
        mostClose = None
        self._actionLock.acquire()

        for tid, tObj in self._tDict.items():
            timeDiff = math.abs(execTimeWanted - self._getTimeInfoAbout(tObj))
            if mostClose is None or mostClose[1] > timeDiff:
                mostClose = (tid, timeDiff)

        self._actionLock.release()

        if mostClose is None or (maxDiff >= 0 and mostClose[1] > maxDiff):
            return None
        else:
            return self.getSpecificTask(mostClose[0])

    def getTasksIDsWithExecTime(self, execTimeWanted, maxDiff= -1.):
        # Return task list containing tasks which the durations approximately
        # sum to execTimeWanted
        returnList = []
        totalTime = 0.
        self._actionLock.acquire()

        for tid, tObj in self._tDict.items():
            timeInfo = self._getTimeInfoAbout(tObj)
            if totalTime + timeInfo <= execTimeWanted:
                returnList.append(tid)
                totalTime += timeInfo
        #
        # Hack to avoid the starting bug (few jobs but very long)
        # 
        if len(returnList) == 0 and maxDiff == -1 and len(self._tDict) > 1:
            it = list(filter(lambda x: self._getTimeInfoAbout(x[1]) != 0, self._tDict.items()))
            if len(it) != 0:
                returnList.append(it[0][0])
                totalTime += self._getTimeInfoAbout(it[0][1])
        self._actionLock.release()
        return returnList, totalTime



class Control(object):
    """
    Control is the main DTM class. The dtm object you receive when you use ``from deap import dtm``
    is a proxy over an instance of this class.

    Most of its methods are used by your program, in the execution tasks; however, two of thems (start() and setOptions()) MUST be called
    in the MainThread (i.e. the thread started by the Python interpreter).

    As this class is instancied directly in the module, initializer takes no arguments.
    """
    def __init__(self):
        self.sTime = time.time()
        
        # Key : Target, Value : StatsContainer
        self.tasksStats = {}
        self.tasksStatsLock = threading.Lock()

        # Global execution lock
        self.dtmExecLock = ExecInfo(self.tasksStats, self.tasksStatsLock)

        self.waitingThreadsQueue = TaskQueue(self.tasksStats, self.tasksStatsLock)
        
        # Key : task ID, Value : WaitInfoContainer
        self.waitingThreads = {}
        self.waitingThreadsLock = threading.Lock()


        self.launchTaskLock = threading.Lock()

        self.waitingForRestartQueue = TaskQueue(self.tasksStats, self.tasksStatsLock)
        self.execQueue = TaskQueue(self.tasksStats, self.tasksStatsLock)

        self.recvQueue = Queue.Queue()      # Contains MessageContainer
        self.sendQueue = Queue.Queue()      # Contains MessageContainer

        self.exitStatus = threading.Event()
        self.commExitNotification = threading.Event()
        self.commReadyEvent = threading.Event()

        self.exitState = (None, None)
        self.exitSetHere = False
        
        # Will stop the main thread until an event occurs
        self.runningFlag = threading.Event()
        self.runningFlag.set()

        self.commManagerType = "deap.dtm.commManagerMpi4py"
        self.loadBalancerType = "deap.dtm.loadBalancerPDB"
        self.printExecSummary = True
        
        self.isStarted = False
        
        self.traceMode = False
        self.traceRoot = None   # Root element of the XML log
        self.traceTasks = None  # XML elements
        self.traceComm = None   # defined if traceMode == True
        self.traceLoadB = None
        self.traceLock = threading.Lock()
        
        self.refTime = 1.

        self.loadBalancer = None

        self.lastRetValue = None
        
        self.dtmRandom = random.Random()



    def _doCleanUp(self):
        """
        Clean up function, called at this end of the execution.
        Should NOT be called by the user.
        """
        #if self.printExecSummary:
             #_logger.info("[%s] did %i tasks", str(self.workerId), self._DEBUG_COUNT_EXEC_TASKS)

        if self.exitSetHere:
            for n in self.commThread.iterOverIDs():
                if n == self.workerId:
                    continue
                self.sendQueue.put(MessageContainer(msgType = DTM_MSG_EXIT,
                                    senderWid = self.workerId,
                                    receiverWid = n,
                                    loadsDict = None,
                                    targetsStats = None,
                                    prepTime = time.time(),
                                    sendTime = 0,
                                    ackNbr = -1,
                                    msg = (0, "Exit with success")))

        self.commExitNotification.set()
        self.commThread.join()

        del self.execQueue
        del self.sendQueue
        del self.recvQueue

        countThreads = sum([1 for th in threading.enumerate() if not th.daemon])
        if countThreads > 1:
            _logger.warning("[%s] There's more than 1 active thread (%i) at the exit.", str(self.workerId), threading.activeCount())
        
        
        if self.commThread.isRootWorker:
            if self.printExecSummary:
                msgT = " Tasks execution times summary :\n"
                for target, times in self.tasksStats.items():
                    msgT += "\t" + str(target) + " : Avg=" + str(times.rAvg * self.refTime) + ", StdDev=" + str(times.rStdDev * self.refTime) + "\n\n"
                _logger.info(msgT)

            _logger.info("DTM execution ended (no more tasks)")
            _logger.info("Total execution time : %s", str(time.time() - self.sTime))
        
        if self.traceMode:
            _logger.info("Writing log to the log files...")
            self.traceLock.acquire()
            fstat = etree.SubElement(self.traceRoot, "finalStats")
            for target, times in self.tasksStats.items():
                etree.SubElement(fstat, "taskinfo", {"target" : str(target), "avg" : repr(times.rAvg * self.refTime), "stddev" : repr(times.rStdDev * self.refTime), "execcount" : str(times.execCount)})
            
            flog = open(DTM_LOGDIR_DEFAULT_NAME + "/log" + str(self.workerId) + ".xml", 'w')
            if PRETTY_PRINT_SUPPORT:
                flog.write(etree.tostring(self.traceRoot, pretty_print=True))
            else:
                flog.write(etree.tostring(self.traceRoot))
            flog.close()
            self.traceLock.release()
            
        if self.exitSetHere:
            if self.lastRetValue[0]:
                return self.lastRetValue[1]
            else:
                raise self.lastRetValue[1]


    def _addTaskStat(self, taskKey, timeExec):
        # The execution time is based on the calibration ref time
        comparableLoad = timeExec / self.refTime
        # We do not keep up all the execution times, but
        # update the mean and stddev realtime
        
        self.tasksStatsLock.acquire()
        if not taskKey in self.tasksStats:
            self.tasksStats[taskKey] = StatsContainer(rAvg = timeExec,
                                                rStdDev = 0.,
                                                rSquareSum = timeExec * timeExec,
                                                execCount = 1)
        else:
            oldAvg = self.tasksStats[taskKey].rAvg
            oldStdDev = self.tasksStats[taskKey].rStdDev
            oldSum2 = self.tasksStats[taskKey].rSquareSum
            oldExecCount = self.tasksStats[taskKey].execCount

            self.tasksStats[taskKey].rAvg = (timeExec + oldAvg * oldExecCount) / (oldExecCount + 1)
            self.tasksStats[taskKey].rSquareSum = oldSum2 + timeExec * timeExec
            self.tasksStats[taskKey].rStdDev = abs(self.tasksStats[taskKey].rSquareSum / (oldExecCount + 1) - self.tasksStats[taskKey].rAvg ** 2) ** 0.5
            self.tasksStats[taskKey].execCount = oldExecCount + 1

        self.tasksStatsLock.release()


    def _calibrateExecTime(self, runsN=3):
        """
        Small calibration test, run at startup
        Should not be called by user
        """
        timesList = []
        for r in range(runsN):
            timeS = time.clock()

            a = 0.
            for i in range(10000):
                a = math.sqrt(self.dtmRandom.random() / (self.dtmRandom.uniform(0, i) + 1))

            strT = ""
            for i in range(5000):
                strT += str(self.dtmRandom.randint(0, 9999))

            for i in range(500):
                pickStr = pickle.dumps(strT)
                strT = pickle.loads(pickStr)

            timesList.append(time.clock() - timeS)

        return sorted(timesList)[int(runsN / 2)]


    def _getLoadTuple(self):
        return (self.dtmExecLock.getLoad(), self.execQueue.getLoad(), self.waitingForRestartQueue.getLoad(), self.waitingThreadsQueue.getLoad())


    def _returnResult(self, idToReturn, resultInfo):
        """
        Called by the execution threads when they have to return a result
        Should NOT be called explicitly by the user
        """
        if idToReturn == self.workerId:
            self._dispatchResults((resultInfo,))
        else:
            self.sendQueue.put(MessageContainer(msgType = DTM_MSG_RESULT,
                                            senderWid = self.workerId,
                                            receiverWid = idToReturn,
                                            loadsDict = self.loadBalancer.getNodesDict(),
                                            targetsStats = self.tasksStats,
                                            prepTime = time.time(),
                                            sendTime = 0,
                                            ackNbr = -1,
                                            msg = (resultInfo,)))

    def _updateStats(self, msg):
        """
        Called by the control thread to update its dictionnary
        Should NOT be called explicitly by the user
        """
        self.loadBalancer.mergeNodeStatus(msg.loadsDict)

        self.tasksStatsLock.acquire()

        for key, val in msg.targetsStats.items():
            if not key in self.tasksStats or val.execCount > self.tasksStats[key].execCount:
                self.tasksStats[key] = val

        self.tasksStatsLock.release()


    def _dispatchResults(self, resultsList):
        """
        Called by the control thread when a message is received;
        Dispatch it to the task waiting for it.
        Should NOT be called explicitly by the user
        """
        for result in resultsList:
            self.waitingThreadsLock.acquire()
            
            # We look for the task waiting for each result
            foundKey = None
            for taskKey in self.waitingThreads[result.parentTid].rWaitingDict:
                try:
                    self.waitingThreads[result.parentTid].rWaitingDict[taskKey].tids.remove(result.tid)
                except ValueError:
                    # The ID is not in this waiting list, continue with other worker
                    continue
                
                foundKey = taskKey
                if isinstance(self.waitingThreads[result.parentTid].rWaitingDict[taskKey].result, list):
                    if not result.success:
                        # Exception occured
                        self.waitingThreads[result.parentTid].rWaitingDict[taskKey].result = result.result
                    else:
                        self.waitingThreads[result.parentTid].rWaitingDict[taskKey].result[result.taskIndex] = result.result
                break
            
            assert not foundKey is None, "Parent task not found for result dispatch!"       # Debug

            if len(self.waitingThreads[result.parentTid].rWaitingDict[foundKey].tids) == 0:
                # All tasks done
                self.waitingThreads[result.parentTid].rWaitingDict[foundKey].finished = True
                if isinstance(self.waitingThreads[result.parentTid].rWaitingDict[foundKey].result, list):
                    self.waitingThreads[result.parentTid].rWaitingDict[foundKey].success = True
                
                canRestart = False
                if self.waitingThreads[result.parentTid].rWaitingDict[foundKey].waitingOn == True or self.waitingThreads[result.parentTid].waitingMode == DTM_WAIT_ALL:
                    if (self.waitingThreads[result.parentTid].waitingMode == DTM_WAIT_ALL and len(self.waitingThreads[result.parentTid].rWaitingDict) == 1)\
                      or self.waitingThreads[result.parentTid].waitingMode == DTM_WAIT_ANY:
                        canRestart = True
                    elif self.waitingThreads[result.parentTid].waitingMode == DTM_WAIT_SOME:
                        canRestart = True
                        
                        for rKey in self.waitingThreads[result.parentTid].rWaitingDict: 
                            if self.waitingThreads[result.parentTid].rWaitingDict[rKey].waitingOn and rKey != foundKey:
                                canRestart = False
                
                if not self.waitingThreads[result.parentTid].rWaitingDict[taskKey].callbackFunc is None:
                    self.waitingThreads[result.parentTid].rWaitingDict[taskKey].callbackFunc()
                
                if canRestart:
                    wTask = self.waitingThreadsQueue.getSpecificTask(result.parentTid)
                    assert not wTask is None
                    self.waitingForRestartQueue.put(wTask)
                
            self.waitingThreadsLock.release()
            
        

    def _startNewTask(self):
        """
        Start a new task (if there's one available)
        Return True if so
        Should NOT be called explicitly by the user
        """
        taskLauched = False
        self.launchTaskLock.acquire()
        if not self.dtmExecLock.isLocked():
            try:
                wTask = self.waitingForRestartQueue.getTask()
                self.dtmExecLock.acquire(wTask)
                wTask.threadObject.waitingFlag.set()
                taskLauched = True
            except Queue.Empty:
                pass

            if not taskLauched:
                try:
                    newTask = self.execQueue.getTask()
                    if self.traceMode:
                        self.traceLock.acquire()
                        newTaskElem = etree.SubElement(self.traceTasks, "task",
                                                       {"id" : str(newTask.tid),
                                                        "creatorTid" : str(newTask.creatorTid),
                                                        "creatorWid" : str(newTask.creatorWid),
                                                        "taskIndex" : str(newTask.taskIndex),
                                                        "creationTime" : repr(newTask.creationTime)})
                        try:
                            newTaskTarget = etree.SubElement(newTaskElem, "target",
                                                         {"name" : str(newTask.target.__name__)})
                        except AttributeError:
                            newTaskTarget = etree.SubElement(newTaskElem, "target",
                                                         {"name" : str(newTask.target)})
                            
                        for i, a in enumerate(newTask.args):
                            newTaskTarget.set("arg" + str(i), str(a))
                        for k in newTask.kwargs:
                            newTaskTarget.set("kwarg_" + str(k), str(newTask.kwargs[k]))
                        
                        newTaskPath = etree.SubElement(newTaskElem, "path", {"data" : str(newTask.taskRoute)})
                        self.traceLock.release()
                        
                        newThread = DtmThread(newTask, self, newTaskElem)
                    else:
                        newThread = DtmThread(newTask, self)
                    self.dtmExecLock.acquire(newTask)
                    newThread.start()
                    taskLauched = True
                except Queue.Empty:
                    pass
            
        self.launchTaskLock.release()
        return taskLauched
    
    def _main(self):
        """
        Main loop of the control thread
        Should NOT be called explicitly by the user
        """
        
        timeBegin = time.time()
        while True:
            
            self.runningFlag.wait()         # WARNING, may deadlock on very specific conditions
            self.runningFlag.clear()        # (if the _last_ task do a set() between those 2 lines, nothing will wake up the main thread)
            
            while True:
                try:
                    recvMsg = self.recvQueue.get_nowait()
                    if recvMsg.msgType == DTM_MSG_EXIT:
                        self.exitStatus.set()
                        self.exitState = (recvMsg.msg[1], recvMsg.msg[0])
                        break
                    elif recvMsg.msgType == DTM_MSG_TASK:
                        self.execQueue.putList(recvMsg.msg)
                        self.loadBalancer.updateSelfStatus(self._getLoadTuple())
                        self.sendQueue.put(MessageContainer(msgType = DTM_MSG_ACK_RECEIVED_TASK,
                                                        senderWid = self.workerId,
                                                        receiverWid = recvMsg.senderWid,
                                                        loadsDict = self.loadBalancer.getNodesDict(),
                                                        targetsStats = self.tasksStats,
                                                        prepTime = time.time(),
                                                        sendTime = 0,
                                                        ackNbr = -1,
                                                        msg = recvMsg.ackNbr))
                        self._updateStats(recvMsg)
                    elif recvMsg.msgType == DTM_MSG_RESULT:
                        self._dispatchResults(recvMsg.msg)
                        self._updateStats(recvMsg)
                    elif recvMsg.msgType == DTM_MSG_REQUEST_TASK:
                        self._updateStats(recvMsg)
                    elif recvMsg.msgType == DTM_MSG_ACK_RECEIVED_TASK:
                        self.loadBalancer.acked(recvMsg.senderWid, recvMsg.msg)
                        self._updateStats(recvMsg)
                    else:
                        _logger.warning("[%s] Unknown message type %s received will be ignored.", str(self.workerId), str(recvMsg.msgType))
                except Queue.Empty:
                    break

	    
            if self.exitStatus.is_set():
                break
            
            currentNodeStatus = self._getLoadTuple()
            self.loadBalancer.updateSelfStatus(currentNodeStatus)

            sendUpdateList, sendTasksList = self.loadBalancer.takeDecision()
            self.tasksStatsLock.acquire()
            for sendInfo in sendTasksList:
                self.sendQueue.put(MessageContainer(msgType = DTM_MSG_TASK,
                                            senderWid = self.workerId,
                                            receiverWid = sendInfo[0],
                                            loadsDict = self.loadBalancer.getNodesDict(),
                                            targetsStats = self.tasksStats,
                                            prepTime = time.time(),
                                            sendTime = 0,
                                            ackNbr = sendInfo[2],
                                            msg = sendInfo[1]))

            for updateTo in sendUpdateList:
                self.sendQueue.put(DtmMessageContainer(msgType = DTM_MSG_REQUEST_TASK,
                                            senderWid = self.workerId,
                                            receiverWid = updateTo,
                                            loadsDict = self.loadBalancer.getNodesDict(),
                                            targetsStats = self.tasksStats,
                                            prepTime = time.time(),
                                            sendTime = 0,
                                            ackNbr = -1,
                                            msg = None))
            self.tasksStatsLock.release()
            self._startNewTask()
        return self._doCleanUp()


    def setOptions(self, *args, **kwargs):
        """
        Set a DTM global option.
        
        .. warning::
            This function must be called BEFORE ``start()``. It is also the user responsability to ensure that the same option is set on every worker.

        Currently, the supported options are :
            * **communicationManager** : can be *deap.dtm.mpi4py* (default) or *deap.dtm.commManagerTCP*.
            * **loadBalancer** : currently only the default *PDB* is available.
            * **printSummary** : if set, DTM will print a task execution summary at the end (mean execution time of each tasks, how many tasks did each worker do, ...)
            * **setTraceMode** : if set, will enable a special DTM tracemode. In this mode, DTM logs all its activity in XML files (one by worker). Mainly for DTM debug purpose, but can also be used for profiling.
        
        This function can be called more than once. Any unknown parameter will have no effect.
        """
        if self.isStarted:
            if self.commThread.isRootWorker:
                _logger.warning("dtm.setOptions() was called after dtm.start(); options will not be considered")
            return
            
        for opt in kwargs:
            if opt == "communicationManager":
                self.commManagerType = kwargs[opt]
            elif opt == "loadBalancer":
                self.loadBalancerType = kwargs[opt]
            elif opt == "printSummary":
                self.printExecSummary = kwargs[opt]
            elif opt == "setTraceMode":
                self.traceMode = kwargs[opt]
            elif self.commThread.isRootWorker:
                _logger.warning("Unknown option '%s'", opt)


    def start(self, initialTarget, *args, **kwargs):
        """
        Start the execution with the target `initialTarget`.
        Calling this function create and launch the first task on the root worker
        (defined by the communication manager, for instance, with MPI, the root worker is the worker with rank 0.).

        .. warning::
            This function must be called only ONCE, and after the target has been parsed by the Python interpreter.
        """
        self.isStarted = True
       
        try:
            tmpImport = __import__(self.commManagerType, globals(), locals(), ['CommThread'], 0)
            if not hasattr(tmpImport, 'CommThread'):
                raise ImportError
            CommThread = tmpImport.CommThread
        except ImportError:
            _logger.warning("Warning : %s is not a suitable communication manager. Default to commManagerMpi4py.", self.commManagerType)
            tmpImport = __import__('deap.dtm.commManagerMpi4py', globals(), locals(), ['CommThread'], 0)
            CommThread = tmpImport.CommThread

        try:
            tmpImport = __import__(self.loadBalancerType, globals(), locals(), ['LoadBalancer'], 0)
            if not hasattr(tmpImport, 'LoadBalancer'):
                raise ImportError
            LoadBalancer = tmpImport.LoadBalancer
        except ImportError:
            _logger.warning("Warning : %s is not a suitable load balancer. Default to loadBalancerPDB.", self.loadBalancerType)
            tmpImport = __import__('deap.dtm.loadBalancerPDB', globals(), locals(), ['LoadBalancer'], 0)
            LoadBalancer = tmpImport.LoadBalancer
        
        self.commThread = CommThread(self.recvQueue, self.sendQueue, self.runningFlag, self.commExitNotification, self.commReadyEvent, self.dtmRandom, sys.argv)

        self.commThread.start()
        self.refTime = self._calibrateExecTime()

        self.commReadyEvent.wait()
        
        if self.commThread.isLaunchProcess:
            sys.exit()

        self.poolSize = self.commThread.poolSize
        self.workerId = self.commThread.workerId

        self.idGenerator = TaskIdGenerator(self.workerId)

        self.loadBalancer = LoadBalancer(self.commThread.iterOverIDs(), self.workerId, self.execQueue, self.dtmRandom)
        
        if self.traceMode:
            self.traceLock.acquire()
            self.traceRoot = etree.Element("dtm", {"version" : str(0.7), "workerId" : str(self.workerId), "timeBegin" : repr(self.sTime)})
            self.traceTasks = etree.SubElement(self.traceRoot, "tasksLog")
            self.traceComm = etree.SubElement(self.traceRoot, "commLog")
            self.traceLoadB = etree.SubElement(self.traceRoot, "loadBalancerLog")
            self.traceLock.release()
            self.commThread.setTraceModeOn(self.traceComm)
            self.loadBalancer.setTraceModeOn(self.traceLoadB)
        
        if self.commThread.isRootWorker:
            
            if self.traceMode:
                # Create the log folder
                try:
                    os.mkdir(DTM_LOGDIR_DEFAULT_NAME)
                except OSError:
                    _logger.warning("Log folder '" + DTM_LOGDIR_DEFAULT_NAME + "' already exists!")
                        
            
            _logger.info("DTM started with %i workers", self.poolSize)
            _logger.info("DTM load balancer is %s, and communication manager is %s", self.loadBalancerType, self.commManagerType)
            
            initTask = TaskContainer(tid = self.idGenerator.tid,
                                    creatorWid = self.workerId,
                                    creatorTid = None,
                                    taskIndex = 0,
                                    taskRoute = [self.workerId],
                                    creationTime = time.time(),
                                    target = initialTarget,
                                    args = args,
                                    kwargs = kwargs,
                                    threadObject = None,
                                    taskState = DTM_TASK_STATE_IDLE)
            self.execQueue.put(initTask)

        return self._main()



    # The following methods are NOT called by the control thread, but by the EXECUTION THREADS
    # All the non-local objects used MUST be thread-safe

    def map(self, function, *iterables, **kwargs):
        """
        A parallel equivalent of the :func:`map` built-in function. It blocks till the result is ready.
        This method chops the iterables into a number of chunks determined by DTM in order to get the most efficient use of the workers.
        It takes any number of iterables (though it will shrink all of them to the len of the smallest one), and any others kwargs that will be
        transmitted as is to the *function* target.
        """
        cThread = threading.currentThread()
        currentId = cThread.tid
        
        zipIterable = list(zip(*iterables))
        
        listResults = [None] * len(zipIterable)
        listTasks = []
        listTids = []
        
        for index, elem in enumerate(zipIterable):
            task = TaskContainer(tid = self.idGenerator.tid,
                                    creatorWid = self.workerId,
                                    creatorTid = currentId,
                                    taskIndex = index,
                                    taskRoute = [self.workerId],
                                    creationTime = time.time(),
                                    target = function,
                                    args = elem,
                                    kwargs = kwargs,
                                    threadObject = None,
                                    taskState = DTM_TASK_STATE_IDLE)
            listTasks.append(task)
            listTids.append(task.tid)
        
        if self.traceMode:
            self.traceLock.acquire()
            newTaskElem = etree.SubElement(cThread.xmlTrace, "event",
                                           {"type" : "map",
                                            "time" : repr(time.time()),
                                            "target" : str(function.__name__),
                                            "childTasks" : str(listTids)})
            self.traceLock.release()
        
        self.waitingThreadsLock.acquire()        
        if currentId not in self.waitingThreads.keys():
            self.waitingThreads[currentId] = WaitInfoContainer(threadObject = cThread,
                                                    eventObject = cThread.waitingFlag,
                                                    waitBeginningTime = 0,
                                                    tasksWaitingCount = 0,
                                                    waitingMode = DTM_WAIT_NONE,
                                                    rWaitingDict = {})
        
        resultKey = listTids[0]
        self.waitingThreads[currentId].rWaitingDict[resultKey] = ExceptedResultContainer(tids = listTids,
                                waitingOn = True,
                                finished = False,
                                success = False,
                                callbackFunc = None,
                                result = listResults)
                                
        self.waitingThreads[currentId].tasksWaitingCount += len(listTasks)
        self.waitingThreads[currentId].waitingMode = DTM_WAIT_SOME
        
        
        self.waitingThreadsQueue.put(cThread.taskInfo)
        
        self.waitingThreads[currentId].waitBeginningTime = time.time()
        self.waitingThreadsLock.release()
        
        self.execQueue.putList(listTasks)
        
        
        cThread.waitForResult()

        self.waitingThreadsLock.acquire()
        ret = self.waitingThreads[currentId].rWaitingDict[resultKey].result
        if self.waitingThreads[currentId].rWaitingDict[resultKey].success == False:
            # Exception occured
            del self.waitingThreads[currentId].rWaitingDict[resultKey]
            self.waitingThreadsLock.release()
            raise ret
        else:           
            del self.waitingThreads[currentId].rWaitingDict[resultKey]
            self.waitingThreadsLock.release()
            return ret


    def map_async(self, function, iterable, callback=None):
        """
        A non-blocking variant of the :func:`~deap.dtm.taskmanager.Control.map` method which returns a :class:`~deap.dtm.taskmanager.AsyncResult` object.
        
        .. note::
            As on version 0.2, callback is not implemented.
        """
        
        cThread = threading.currentThread()
        currentId = cThread.tid

        listResults = [None] * len(iterable)
        listTasks = []
        listTids = []

        for index, elem in enumerate(iterable):
            task = TaskContainer(tid = self.idGenerator.tid,
                                    creatorWid = self.workerId,
                                    creatorTid = currentId,
                                    taskIndex = index,
                                    taskRoute = [self.workerId],
                                    creationTime = time.time(),
                                    target = function,
                                    args = (elem,),
                                    kwargs = {},
                                    threadObject = None,
                                    taskState = DTM_TASK_STATE_IDLE)
            listTasks.append(task)
            listTids.append(task.tid)
        
        resultKey = listTids[0]
        
        if self.traceMode:
            newTaskElem = etree.SubElement(cThread.xmlTrace, "event",
                                           {"type" : "map_async",
                                            "time" : repr(time.time()),
                                            "target" : str(function.__name__),
                                            "childTasks" : str(listTids)})
        
        self.waitingThreadsLock.acquire()
        
        if currentId not in self.waitingThreads.keys():
            self.waitingThreads[currentId] = WaitInfoContainer(threadObject = cThread,
                                                    eventObject = cThread.waitingFlag,
                                                    waitBeginningTime = 0,
                                                    tasksWaitingCount = 0,
                                                    waitingMode = DTM_WAIT_NONE,
                                                    rWaitingDict = {})
                                                    
        self.waitingThreads[currentId].rWaitingDict[resultKey] = ExceptedResultContainer(tids = listTids,
                                waitingOn = False,
                                finished = False,
                                success = False,
                                callbackFunc = None,
                                result = listResults)

        self.waitingThreads[currentId].waitingMode = DTM_WAIT_NONE
        
        asyncRequest = AsyncResult(self, self.waitingThreads[currentId], resultKey)
        self.waitingThreads[currentId].rWaitingDict[resultKey].callbackFunc = asyncRequest._dtmCallback
        
        self.waitingThreadsLock.release()
        
        self.execQueue.putList(listTasks)
        
        self.runningFlag.set()
        
        return asyncRequest


    def apply(self, function, *args, **kwargs):
        """
        Equivalent of the :func:`apply` built-in function. It blocks till the result is ready.
        Given this blocks, :func:`~deap.dtm.taskmanager.Control.apply_async()` is better suited for performing work in parallel.
        Additionally, the passed in function is only executed in one of the workers of the pool.
        """
        cThread = threading.currentThread()
        currentId = cThread.tid
        
        task = TaskContainer(tid = self.idGenerator.tid,
                                    creatorWid = self.workerId,
                                    creatorTid = currentId,
                                    taskIndex = 0,
                                    taskRoute = [self.workerId],
                                    creationTime = time.time(),
                                    target = function,
                                    args = args,
                                    kwargs = kwargs,
                                    threadObject = None,
                                    taskState = DTM_TASK_STATE_IDLE)
        
        if self.traceMode:
            newTaskElem = etree.SubElement(cThread.xmlTrace, "event",
                                           {"type" : "apply",
                                            "time" : repr(time.time()),
                                            "target" : str(function.__name__),
                                            "childTasks" : str([task.tid])})
        
        self.waitingThreadsLock.acquire()
        if currentId not in self.waitingThreads.keys():
            self.waitingThreads[currentId] = WaitInfoContainer(threadObject = cThread,
                                                    eventObject = cThread.waitingFlag,
                                                    waitBeginningTime = 0,
                                                    tasksWaitingCount = 0,
                                                    waitingMode = DTM_WAIT_NONE,
                                                    rWaitingDict = {})
        
        resultKey = task.tid
        
        self.waitingThreads[currentId].rWaitingDict[resultKey] = ExceptedResultContainer(tids = [task.tid],
                                waitingOn = True,
                                finished = False,
                                success = False,
                                callbackFunc = None,
                                result = [None])
        
        self.waitingThreads[currentId].tasksWaitingCount += 1
        self.waitingThreads[currentId].waitingMode = DTM_WAIT_SOME
        
        self.waitingThreadsQueue.put(cThread.taskInfo)
        
        self.waitingThreads[currentId].waitBeginningTime = time.time()
        self.waitingThreadsLock.release()
        
        self.execQueue.put(task)
        

        cThread.waitForResult()

        self.waitingThreadsLock.acquire()
        ret = self.waitingThreads[currentId].rWaitingDict[resultKey].result
        if self.waitingThreads[currentId].rWaitingDict[resultKey].success == False:
            # Exception occured
            del self.waitingThreads[currentId].rWaitingDict[resultKey]
            self.waitingThreadsLock.release()
            raise ret
        else:           
            del self.waitingThreads[currentId].rWaitingDict[resultKey]
            self.waitingThreadsLock.release()
            return ret[0]

    def apply_async(self, function, *args, **kwargs):
        """
        A non-blocking variant of the :func:`~deap.dtm.taskmanager.Control.apply` method which returns a :class:`~deap.dtm.taskmanager.AsyncResult` object.
        """
        cThread = threading.currentThread()
        currentId = cThread.tid
        
        task = TaskContainer(tid = self.idGenerator.tid,
                                    creatorWid = self.workerId,
                                    creatorTid = currentId,
                                    taskIndex = 0,
                                    taskRoute = [self.workerId],
                                    creationTime = time.time(),
                                    target = function,
                                    args = args,
                                    kwargs = kwargs,
                                    threadObject = None,
                                    taskState = DTM_TASK_STATE_IDLE)
        
        if self.traceMode:
            newTaskElem = etree.SubElement(cThread.xmlTrace, "event",
                                           {"type" : "apply_async",
                                            "time" : repr(time.time()),
                                            "target" : str(function.__name__),
                                            "childTasks" : str([task.tid])})
        
        self.waitingThreadsLock.acquire()        
        if currentId not in self.waitingThreads.keys():
            self.waitingThreads[currentId] = WaitInfoContainer(threadObject = cThread,
                                                    eventObject = cThread.waitingFlag,
                                                    waitBeginningTime = 0,
                                                    tasksWaitingCount = 0,
                                                    waitingMode = DTM_WAIT_NONE,
                                                    rWaitingDict = {})
        
        resultKey = task.tid
        
        self.waitingThreads[currentId].rWaitingDict[resultKey] = ExceptedResultContainer(tids = [task.tid],
                                waitingOn = False,
                                finished = False,
                                success = False,
                                callbackFunc = None,
                                result = [None])
        
        self.waitingThreads[currentId].waitingMode = DTM_WAIT_NONE
        
        asyncRequest = AsyncResult(self, self.waitingThreads[currentId], resultKey)
        self.waitingThreads[currentId].rWaitingDict[resultKey].callbackFunc = asyncRequest._dtmCallback
        
        self.waitingThreadsLock.release()
        
        self.execQueue.put(task)
        
        self.runningFlag.set()
        return asyncRequest

    def imap(self, function, iterable, chunksize=1):
        """
        An equivalent of :func:`itertools.imap`.

        The chunksize argument can be used to tell DTM how many elements should be computed at the same time.
        For very long iterables using a large value for chunksize can make make the job complete much faster than using the default value of 1.
        """
        currentIndex = 0

        while currentIndex < len(iterable):
            maxIndex = currentIndex + chunksize if currentIndex + chunksize < len(iterable) else len(iterable)
            asyncResults = [None] * (maxIndex - currentIndex)
            for i in range(currentIndex, maxIndex):
                asyncResults[i % chunksize] = self.apply_async(function, iterable[i])

            for result in asyncResults:
                ret = result.get()
                yield ret

            currentIndex = maxIndex


    def imap_unordered(self, function, iterable, chunksize=1):
        """
        Not implemented yet.
        """
        raise NotImplementedError


    def filter(self, function, iterable):
        """
        Same behavior as the built-in :func:`filter`. The filtering is done localy, but the computation is distributed.
        """
        results = self.map(function, iterable)
        return [item for result, item in zip(results, iterable) if result]

    def repeat(self, function, n, *args, **kwargs):
        """
        Repeat the function *function* *n* times, with given args and keyworded args.
        Return a list containing the results.
        """
        cThread = threading.currentThread()
        currentId = cThread.tid

        listResults = [None] * n
        listTasks = []
        listTids = []
        
        for index in range(n):
            task = TaskContainer(tid = self.idGenerator.tid,
                                    creatorWid = self.workerId,
                                    creatorTid = currentId,
                                    taskIndex = index,
                                    taskRoute = [self.workerId],
                                    creationTime = time.time(),
                                    target = function,
                                    args = args,
                                    kwargs = kwargs,
                                    threadObject = None,
                                    taskState = DTM_TASK_STATE_IDLE)
            listTasks.append(task)
            listTids.append(task.tid)
        
        self.waitingThreadsLock.acquire()        
        if currentId not in self.waitingThreads.keys():
            self.waitingThreads[currentId] = WaitInfoContainer(threadObject = cThread,
                                                    eventObject = cThread.waitingFlag,
                                                    waitBeginningTime = 0,
                                                    tasksWaitingCount = 0,
                                                    waitingMode = DTM_WAIT_NONE,
                                                    rWaitingDict = {})
        
        resultKey = listTids[0]
        self.waitingThreads[currentId].rWaitingDict[resultKey] = ExceptedResultContainer(tids = listTids,
                                waitingOn = True,
                                finished = False,
                                success = False,
                                callbackFunc = None,
                                result = listResults)
                                
        self.waitingThreads[currentId].tasksWaitingCount += len(listTasks)
        self.waitingThreads[currentId].waitingMode = DTM_WAIT_SOME
        
        
        self.waitingThreadsQueue.put(cThread.taskInfo)
        
        self.waitingThreads[currentId].waitBeginningTime = time.time()
        self.waitingThreadsLock.release()
        
        self.execQueue.putList(listTasks)
        
        
        cThread.waitForResult()

        self.waitingThreadsLock.acquire()
        ret = self.waitingThreads[currentId].rWaitingDict[resultKey].result
        if self.waitingThreads[currentId].rWaitingDict[resultKey].success == False:
            # Exception occured
            del self.waitingThreads[currentId].rWaitingDict[resultKey]
            self.waitingThreadsLock.release()
            raise ret
        else:           
            del self.waitingThreads[currentId].rWaitingDict[resultKey]
            self.waitingThreadsLock.release()
            return ret

    def waitForAll(self):
        """
        Wait for all pending asynchronous results. When this function returns,
        DTM guarantees that all ready() call on asynchronous tasks will
        return true.
        """
        threadId = threading.currentThread().tid
        
        self.waitingThreadsLock.acquire()
        if threadId in self.waitingThreads and len(self.waitingThreads[threadId].rWaitingDict) > 0:
            self.waitingThreads[threadId].waitingMode = DTM_WAIT_ALL
            self.waitingThreadsQueue.put(threading.currentThread().taskInfo)
            self.waitingThreadsLock.release()
            threading.currentThread().waitForResult()
            
            self.waitingThreadsLock.acquire()
            self.waitingThreads[threadId].waitingMode = DTM_WAIT_NONE           
            self.waitingThreadsLock.release()
        else:            
            self.waitingThreadsLock.release()
            return        
        return None

    def testAllAsync(self):
        """
        Check whether all pending asynchronous tasks are done.
        It does not lock if it is not the case, but returns false.
        """
        threadId = threading.currentThread().tid
        self.waitingThreadsLock.acquire()
        if threadId in self.waitingThreads:
            ret = len(self.waitingThreads[threadId].rWaitingDict)
            self.waitingThreadsLock.release()
            return False
        else:
            self.waitingThreadsLock.release()
            return True


    def getWorkerId(self):
        """
        Return a unique ID for the current worker. Depending of the
        communication manager type, it can be virtually any Python
        immutable type.
        
        .. note::
            With MPI, the value returned is the MPI slot number.
        """
        return self.workerId




class DtmThread(threading.Thread):
    """
    DTM execution threads. Those are one of the main parts of DTM.
    They should not be created or called directly by the user.
    """
    def __init__(self, structInfo, controlThread, xmlTrace=None):
        threading.Thread.__init__(self)
        
        self.taskInfo = structInfo      # TaskContainer
        
        self.taskInfo.threadObject = self   # Remind that we are the exec thread
        
        self.tid = structInfo.tid
        self.t = structInfo.target
        self.control = controlThread

        self.waitingFlag = threading.Event()        
        self.waitingFlag.clear()
        
        self.timeExec = 0
        self.timeBegin = 0
        if structInfo.creatorTid is None:
            self.isRootTask = True
        else:
            self.isRootTask = False
        
        self.xmlTrace = xmlTrace

    def run(self):
        # The lock is already acquired for us
        self.taskInfo.taskState = DTM_TASK_STATE_RUNNING
        success = True
        self.timeBegin = time.time()
        if not self.xmlTrace is None:
            # Debug output in xml object
            self.control.traceLock.acquire()
            etree.SubElement(self.xmlTrace, "event", {"type" : "begin", "worker" : str(self.control.workerId), "time" : repr(self.timeBegin)})
            self.control.traceLock.release()
            
        try:
            returnedR = self.t(*self.taskInfo.args, **self.taskInfo.kwargs)
        except Exception as expc:
            returnedR = expc
            strWarn = "An exception of type " + str(type(expc)) + " occured on worker " + str(self.control.workerId) + " while processing task " + str(self.tid)
            _logger.warning(strWarn)
            _logger.warning("This will be transfered to the parent task.")
            _logger.warning("Exception details : " + str(expc))
            success = False
            
        
        self.timeExec += time.time() - self.timeBegin
        
        self.control.dtmExecLock.release()
        
        if not self.xmlTrace is None:
            # Debug output in xml object
            self.control.traceLock.acquire()
            etree.SubElement(self.xmlTrace, "event", {"type" : "end", "worker" : str(self.control.workerId), "time" : repr(time.time()), "execTime" : repr(self.timeExec), "success" : str(success)})
            self.control.traceLock.release()
        
        if success:
            try:
                self.control._addTaskStat(self.t.__name__, self.timeExec)
            except AttributeError:
                self.control._addTaskStat(str(self.t), self.timeExec)

        
        if self.isRootTask:
            # Is this task the root task (launch by dtm.start)? If so, we quit
            self.control.lastRetValue = (success, returnedR)
            self.control.exitSetHere = True
            self.control.exitStatus.set()
        else:
            # Else, tell the communication thread to return the result
            resultStruct = ResultContainer(tid = self.tid,
                                            parentTid = self.taskInfo.creatorTid,
                                            taskIndex = self.taskInfo.taskIndex,
                                            execTime = self.timeExec,
                                            success = success,
                                            result = returnedR)
                                            
            self.control._returnResult(self.taskInfo.creatorWid, resultStruct)

        # Tell the control thread that something happened
        self.control._startNewTask()
        
        if self.isRootTask:
            self.control.runningFlag.set()
        
        self.control.waitingThreadsLock.acquire()
        if self.tid in self.control.waitingThreads.keys():
            del self.control.waitingThreads[self.tid]
        self.control.waitingThreadsLock.release()
        self.taskInfo.taskState = DTM_TASK_STATE_FINISH
    

    def waitForResult(self):
        # Clear the execution lock, and sleep
        beginTimeWait = time.time()
        self.timeExec += beginTimeWait - self.timeBegin
        self.control.dtmExecLock.release()
        
        self.taskInfo.taskState = DTM_TASK_STATE_WAITING
        
        if not self.xmlTrace is None:
            # Debug output in xml object
            self.control.traceLock.acquire()
            etree.SubElement(self.xmlTrace, "event", {"type" : "sleep", "worker" : str(self.control.workerId), "time" : repr(beginTimeWait)})
            self.control.traceLock.release()
        
        self.control._startNewTask()
        self.control.runningFlag.set()
        
        self.waitingFlag.wait()        
        self.waitingFlag.clear()

        # At this point, we already have acquired the execution lock
        self.taskInfo.taskState = DTM_TASK_STATE_RUNNING
        self.timeBegin = time.time()
        
        if not self.xmlTrace is None:
            # Debug output in xml object
            self.control.traceLock.acquire()
            etree.SubElement(self.xmlTrace, "event", {"type" : "wakeUp", "worker" : str(self.control.workerId), "time" : repr(self.timeBegin)})
            self.control.traceLock.release()



class AsyncResult(object):
    """
    The class of the result returned by :func:`~deap.dtm.taskmanager.Control.map_async()` and :func:`~deap.dtm.taskmanager.Control.apply_async()`.
    """
    def __init__(self, control, waitingInfo, taskKey):
        self.control = control
        self.resultReturned = False
        self.resultSuccess = False
        self.resultVal = None
        self.taskKey = taskKey
        self.dictWaitingInfo = waitingInfo
        
    
    def _dtmCallback(self):
        # Used by DTM to inform the object that the job is done
        self.resultSuccess = self.dictWaitingInfo.rWaitingDict[self.taskKey].success
        self.resultVal = self.dictWaitingInfo.rWaitingDict[self.taskKey].result
        self.resultReturned = True
        
        del self.dictWaitingInfo.rWaitingDict[self.taskKey]


    def get(self):
        """
        Return the result when it arrives.
        
        .. note::
            This is a blocking call : caller will wait in this function until the result is ready.
            To check for the avaibility of the result, use :func:`~deap.dtm.taskmanager.AsyncResult.ready()`.
        """
        if not self.resultReturned:
            self.wait()
        
        if self.resultSuccess:
            return self.resultVal
        else:
            raise self.resultVal

    def wait(self):
        """
        Wait until the result is available
        """
        
        self.control.waitingThreadsLock.acquire()
        
        if self.ready():
            # This test MUST be protected by the mutex on waitingThreads
            self.control.waitingThreadsLock.release()
            return
        
        self.control.waitingThreads[threading.currentThread().tid].waitingMode = DTM_WAIT_SOME
        self.control.waitingThreads[threading.currentThread().tid].rWaitingDict[self.taskKey].waitingOn = True
        #self.dictWaitingInfo.waitingMode = DTM_WAIT_SOME
        #self.dictWaitingInfo.rWaitingDict[self.taskKey].waitingOn = True
        self.control.waitingThreadsQueue.put(threading.currentThread().taskInfo)
        self.control.waitingThreadsLock.release()

        threading.currentThread().waitForResult()


    def ready(self):
        """
        Return whether the asynchronous task has completed.
        """
        return self.resultReturned

    def successful(self):
        """
        Return whether the task completed without error. Will raise AssertionError if the result is not ready.
        """
        if not self.resultReturned:
            raise AssertionError("Call AsyncResult.successful() before the results were ready!")
        return self.resultSuccess



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

"""
Distributed computation of the prime numbers below 45,000 using DTM.
"""

import time
import math

from deap import dtm

def primaryTest(nbr):
    if nbr <= 2:
        return True
    
    for i in range(2, int(math.sqrt(nbr))+1):
        if nbr % i == 0:
            return False
    return True

def main(upTo):
    beginTime = time.time()
    listNbr = range(3,upTo,2)
    print("Computation begin...")
    listPrimes = dtm.filter(primaryTest, listNbr)
    
    print("Found " + str(len(listPrimes)) + " prime numbers between 3 and " + str(upTo))
    print("Computation time : " + str(time.time()-beginTime))
	
dtm.start(main, 45003)



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

"""
This example is from "John R. Koza. Genetic Programming: On the Programming 
of Computers by Natural Selection. MIT Press, Cambridge, MA, USA, 1992.".

The problem is called The Artificial Ant Problem. 
<http://www.cs.ucl.ac.uk/staff/w.langdon/bloat_csrp-97-29/node2.html>

The goal of this example is to show how to use DEAP and its GP framework with
with complex system of functions and object. 

Given an AntSimulator ant, this solution should get the 89 pieces of food
within 543 moves.
ant.routine = ant.if_food_ahead(ant.move_forward, prog3(ant.turn_left, 
                                                  prog2(ant.if_food_ahead(ant.move_forward, ant.turn_right), 
                                                        prog2(ant.turn_right, prog2(ant.turn_left, ant.turn_right))),
                                                  prog2(ant.if_food_ahead(ant.move_forward, ant.turn_left), ant.move_forward)))

Best solution found with DEAP:
prog3(prog3(move_forward, 
            turn_right, 
            if_food_ahead(if_food_ahead(prog3(move_forward,
                                              move_forward, 
                                              move_forward), 
                                        prog2(turn_left, 
                                              turn_right)), 
                          turn_left)), 
      if_food_ahead(turn_left, 
                    turn_left), 
      if_food_ahead(move_forward, 
                    turn_right)) 
fitness = (89,)
"""

import sys
import logging
import copy
import random
from functools import partial

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap import gp


logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

def progn(*args):
    for arg in args:
        arg()

def prog2(out1, out2): 
    return partial(progn,out1,out2)

def prog3(out1, out2, out3):     
    return partial(progn,out1,out2,out3)  

def if_then_else(condition, out1, out2):
    out1() if condition() else out2()

class AntSimulator(object):
    direction = ["north","east","south","west"]
    dir_row = [1, 0, -1, 0]
    dir_col = [0, 1, 0, -1]
    
    def __init__(self, max_moves):
        self.max_moves = max_moves
        self.moves = 0
        self.eaten = 0
        self.routine = None
        
    def _reset(self):
        self.row = self.row_start 
        self.col = self.col_start 
        self.dir = 1
        self.moves = 0  
        self.eaten = 0
        self.matrix_exc = copy.deepcopy(self.matrix)

    @property
    def position(self):
        return (self.row, self.col, self.direction[self.dir])
            
    def turn_left(self): 
        if self.moves < self.max_moves:
            self.moves += 1
            self.dir = (self.dir - 1) % 4

    def turn_right(self):
        if self.moves < self.max_moves:
            self.moves += 1    
            self.dir = (self.dir + 1) % 4
        
    def move_forward(self):
        if self.moves < self.max_moves:
            self.moves += 1
            self.row = (self.row + self.dir_row[self.dir]) % self.matrix_row
            self.col = (self.col + self.dir_col[self.dir]) % self.matrix_col
            if self.matrix_exc[self.row][self.col] == "food":
                self.eaten += 1
            self.matrix_exc[self.row][self.col] = "passed"

    def sense_food(self):
        ahead_row = (self.row + self.dir_row[self.dir]) % self.matrix_row
        ahead_col = (self.col + self.dir_col[self.dir]) % self.matrix_col        
        return self.matrix_exc[ahead_row][ahead_col] == "food"
   
    def if_food_ahead(self, out1, out2):
        return partial(if_then_else, self.sense_food, out1, out2)
   
    def run(self,routine):
        self._reset()
        while self.moves < self.max_moves:
            routine()
    
    def parse_matrix(self, matrix):
        self.matrix = list()
        for i, line in enumerate(matrix):
            self.matrix.append(list())
            for j, col in enumerate(line):
                if col == "#":
                    self.matrix[-1].append("food")
                elif col == ".":
                    self.matrix[-1].append("empty")
                elif col == "S":
                    self.matrix[-1].append("empty")
                    self.row_start = self.row = i
                    self.col_start = self.col = j
                    self.dir = 1
        self.matrix_row = len(self.matrix)
        self.matrix_col = len(self.matrix[0])
        self.matrix_exc = copy.deepcopy(self.matrix)

ant = AntSimulator(600)

pset = gp.PrimitiveSet("MAIN", 0)
pset.addPrimitive(ant.if_food_ahead, 2)
pset.addPrimitive(prog2, 2)
pset.addPrimitive(prog3, 3)
pset.addTerminal(ant.move_forward)
pset.addTerminal(ant.turn_left)
pset.addTerminal(ant.turn_right)

creator.create("FitnessMax", base.Fitness, weights=(1.0,))
creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMax, pset=pset)

toolbox = base.Toolbox()

# Attribute generator
toolbox.register("expr_init", gp.genFull, pset=pset, min_=1, max_=2)

# Structure initializers
toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr_init)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

def evalArtificialAnt(individual):
    # Transform the tree expression to functionnal Python code
    routine = gp.evaluate(individual, pset)
    # Run the generated routine
    ant.run(routine)
    return ant.eaten,

toolbox.register("evaluate", evalArtificialAnt)
toolbox.register("select", tools.selTournament, tournsize=7)
toolbox.register("mate", gp.cxUniformOnePoint)
toolbox.register("expr_mut", gp.genFull, min_=0, max_=2)
toolbox.register("mutate", gp.mutUniform, expr=toolbox.expr_mut)

def main():
    random.seed(69)

    trail_file = open("santafe_trail.txt")
    ant.parse_matrix(trail_file)
    
    pop = toolbox.population(n=300)
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)
    
    algorithms.eaSimple(toolbox, pop, 0.5, 0.2, 40, stats, halloffame=hof)
    
    logging.info("Best individual is %s, %s", gp.evaluate(hof[0]), hof[0].fitness)
    
    return pop, hof, stats

if __name__ == "__main__":
    main()




#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

"""
DEAP (Distributed Evolutionary Algorithms in Python) is a novel 
evolutionary computation framework for rapid prototyping and testing of 
ideas. Its design departs from most other existing frameworks in that it 
seeks to make algorithms explicit and data structures transparent, as 
opposed to the more common black box type of frameworks. It also 
incorporates easy parallelism where users need not concern themselves with 
gory implementation details like synchronization and load balancing, only 
functional decomposition.

The five founding hypotheses of DEAP are:

(1) The user knows best. Users should be able to understand the internal 
    mechanisms of the framework so that they can extend them easily to better 
    suit their specific needs.

(2) User needs in terms of algorithms and operators are so vast that it would 
    be unrealistic to think of implementing them all in a single framework. 
    However, it should be possible to build basic tools and generic mechanisms 
    that enable easy user implementation of most any EA variant.

(3) Speedy prototyping of ideas is often more precious than speedy execution 
    of programs. Moreover, code compactness and clarity is also very precious.

(4) Even though interpreted, Python is fast enough to execute EAs. Whenever 
    execution time becomes critical, compute intensive components can always 
    be recoded in C. Many efficient numerical libraries are already available 
    through Python APIs.

(5) Easy parallelism can alleviate slow execution.

And these hypotheses lead to the following objectives:

**Rapid prototyping**
    Provide an environment allowing users to quickly implement their own 
    algorithms without compromise.

**Parallelization made easy**
   Allow for straightforward parallelization; users should not be forced to 
   specify more than the granularity level of their functional decomposition.

**Adaptive load balancing**
    The workload should automatically and dynamically be distributed among 
    available compute units; user intervention should be optional and limited 
    to hints of relative loads of tasks.

**Preach by examples**
    Although the aim of the framework is not to provide ready made solutions, 
    it should nevertheless come with a substantial set of real-world examples 
    to guide the apprenticeship of users.
"""

__author__ = "Francois-Michel De Rainville and Felix-Antoine Fortin"
__version__ = "0.7"
__revision__ = "0.7.1"



#!/usr/bin/env python2.7
#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import csv
import logging
import random
import sys

import knn
from deap import algorithms
from deap import base
from deap import creator
from deap import tools

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

# kNN parameters
import knn
FILE="heart_scale.csv"
N_TRAIN=175
K=1

# Read data from file
data = csv.reader(open(FILE, "rb"))
trainset = list()
trainlabels = list()
rows = [row for row in data]
random.shuffle(rows)
for row in rows:
    trainlabels.append(float(row[0]))
    trainset.append([float(e) for e in row[1:]])

classifier = knn.KNN(K)
classifier.train(trainset[:N_TRAIN], trainlabels[:N_TRAIN])

def evalClassifier(individual):
    labels = classifier.predict(trainset[N_TRAIN:], individual)
    return sum(1 for x, y in zip(labels, trainlabels) if x == y)  / float(len(trainlabels)), \
           sum(individual) / float(classifier.ndim)

creator.create("FitnessMulti", base.Fitness, weights=(1.0, -1.0))
creator.create("Individual", list, fitness=creator.FitnessMulti)

toolbox = base.Toolbox()
# Attribute generator
toolbox.register("attr_bool", random.randint, 0, 1)
# Structure initializers
toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_bool, classifier.ndim)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

# Operator registering
toolbox.register("evaluate", evalClassifier)
toolbox.register("mate", tools.cxUniform, indpb=0.1)
toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
toolbox.register("select", tools.selNSGA2)


def main():
    # random.seed(64)
    MU, LAMBDA = 100, 200
    pop = toolbox.population(n=MU)
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)

    algorithms.eaMuPlusLambda(toolbox, pop, mu=MU, lambda_=LAMBDA,
                              cxpb=0.5, mutpb=0.2, ngen=40, 
                              stats=stats, halloffame=hof)
    logging.info("Best individual is %s, %s", hof[0], hof[0].fitness.values)
    
    return pop, stats, hof

if __name__ == "__main__":
    main()



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import random

from deap import algorithms
from deap import base
from deap import creator
from deap import tools

import sortingnetwork as sn
INPUTS = 12

def evalNetwork(host, parasite, dimension):
    network = sn.SortingNetwork(dimension, host)
    return network.assess(parasite),

def genWire(dimension):
    return (random.randrange(dimension), random.randrange(dimension))

def genNetwork(dimension, min_size, max_size):
    size = random.randint(min_size, max_size)
    return [genWire(dimension) for i in xrange(size)]

def getParasite(dimension):
    return [random.choice((0, 1)) for i in range(dimension)]

def mutNetwork(individual, dimension, mutpb, addpb, delpb, indpb):
    if random.random() < mutpb:
        for index, elem in enumerate(individual):
            if random.random() < indpb:
                individual[index] = genWire(dimension)
    if random.random() < addpb:
        index = random.randint(0, len(individual))
        individual.insert(index, genWire(dimension))
    if random.random() < delpb:
        index = random.randrange(len(individual))
        del individual[index]
    return individual

def mutParasite(individual, indmut, indpb):
    for i in individual:
        if random.random() < indpb:
            indmut(i)
    return individual

creator.create("FitnessMax", base.Fitness, weights=(1.0,))
creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
creator.create("Host", list, fitness=creator.FitnessMin)
creator.create("Parasite", list, fitness=creator.FitnessMax)

htoolbox = base.Toolbox()
ptoolbox = base.Toolbox()

htoolbox.register("network", genNetwork, dimension=INPUTS, min_size=9, max_size=12)
htoolbox.register("individual", tools.initIterate, creator.Host, htoolbox.network)
htoolbox.register("population", tools.initRepeat, list, htoolbox.individual)

ptoolbox.register("parasite", getParasite, dimension=INPUTS)
ptoolbox.register("individual", tools.initRepeat, creator.Parasite, ptoolbox.parasite, 20)
ptoolbox.register("population", tools.initRepeat, list, ptoolbox.individual)

htoolbox.register("evaluate", evalNetwork, dimension=INPUTS)
htoolbox.register("mate", tools.cxTwoPoints)
htoolbox.register("mutate", mutNetwork, dimension=INPUTS, mutpb=0.2, addpb=0.01, 
    delpb=0.01, indpb=0.05)
htoolbox.register("select", tools.selTournament, tournsize=3)

ptoolbox.register("mate", tools.cxTwoPoints)
ptoolbox.register("indMutate", tools.mutFlipBit, indpb=0.05)
ptoolbox.register("mutate", mutParasite, indmut=ptoolbox.indMutate, indpb=0.05)
ptoolbox.register("select", tools.selTournament, tournsize=3)

def cloneHost(individual):
    """Specialized copy function that will work only on a list of tuples
    with no other member than a fitness.
    """
    clone = individual.__class__(individual)
    clone.fitness.values = individual.fitness.values
    return clone

def cloneParasite(individual):
    """Specialized copy function that will work only on a list of lists
    with no other member than a fitness.
    """
    clone = individual.__class__(list(seq) for seq in individual)
    clone.fitness.values = individual.fitness.values
    return clone

htoolbox.register("clone", cloneHost)
ptoolbox.register("clone", cloneParasite)

def main():
    random.seed(64)
    
    hosts = htoolbox.population(n=300)
    parasites = ptoolbox.population(n=300)
    hof = tools.HallOfFame(1)
    hstats = tools.Statistics(lambda ind: ind.fitness.values)
    hstats.register("Avg", tools.mean)
    hstats.register("Std", tools.std)
    hstats.register("Min", min)
    hstats.register("Max", max)
    
    MAXGEN = 50
    H_CXPB, H_MUTPB = 0.5, 0.3
    P_CXPB, P_MUTPB = 0.5, 0.3
    
    fits = htoolbox.map(htoolbox.evaluate, hosts, parasites)
    for host, parasite, fit in zip(hosts, parasites, fits):
        host.fitness.values = parasite.fitness.values = fit
    
    hof.update(hosts)
    hstats.update(hosts)
    
    for g in range(MAXGEN):
        print "-- Generation %i --" % g
        hosts = htoolbox.select(hosts, len(hosts))
        hosts = [htoolbox.clone(ind) for ind in hosts]
        parasites = ptoolbox.select(parasites, len(parasites))
        parasites = [ptoolbox.clone(ind) for ind in parasites]
        
        hosts = algorithms.varSimple(htoolbox, hosts, H_CXPB, H_MUTPB)
        parasites = algorithms.varSimple(ptoolbox, parasites, P_CXPB, P_MUTPB)
        
        fits = htoolbox.map(htoolbox.evaluate, hosts, parasites)
        for host, parasite, fit in zip(hosts, parasites, fits):
            host.fitness.values = parasite.fitness.values = fit
        
        hof.update(hosts)
        hstats.update(hosts)
        
        print hstats
    
    best_network = sn.SortingNetwork(INPUTS, hof[0])
    print best_network
    print best_network.draw()
    print "%i errors" % best_network.assess()

    return hosts, hstats, hof

if __name__ == "__main__":
    main()


#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

"""
DTM version of the sssched utility by Devert and Gagne
(http://marmakoide.org/content/code/sssched.html)

Execute concurrently a list of independant tasks
specified in a file passed with the -c command line option.
"""


import subprocess
import time
import sys
import optparse
import logging


from deap import dtm

_logger = logging.getLogger("dtm.sssched")

def executor(taskCmd):
    _logger.info("Lauching %s on worker %s", taskCmd, dtm.getWorkerId())
    return subprocess.call(taskCmd, stdout=sys.stdout, stderr=sys.stderr)

def main(fTasks):
    tasksList = []
    with open(fTasks) as f:
        for line in f:
            line = line.strip()
            if len(line) > 0:
                tasksList.append(line)
    
    retVals = dtm.map(executor, tasksList)
    if sum(retVals) > 0:
        _logger.warning("Some tasks did not run well (return code > 0) :")
        for i in xrange(len(retVals)):
            if retVals[i] > 0:
                _logger.warning("%s returns %s", tasksList[i], retVals[i])
    
    _logger.info("Done")

if __name__ == '__main__':
    cmdLine = optparse.OptionParser(description = "Simple script scheduler (sssched) for remote job management")
    cmdLine.add_option("-c", "--commands", 
                        dest = "commandsFileName",
                        action = "store",
                        help = "File with list of commands to perform (default:\"commands.lst\")",
                        type = "string")
    
    (options, args) = cmdLine.parse_args()
    dtm.start(main, options.commandsFileName)


#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import array
import logging
import sys
import random

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

from deap import algorithms
from deap import base
from deap import benchmarks
from deap import creator
from deap import tools

creator.create("FitnessMin", base.Fitness, weights=(-1.0, -1.0))
creator.create("Individual", array.array, typecode='d', fitness=creator.FitnessMin)

toolbox = base.Toolbox()

# Attribute generator
toolbox.register("attr_float", random.uniform, -5, 5)

# Structure initializers
toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_float, 3)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

def checkBounds(min, max):
    def decCheckBounds(func):
        def wrapCheckBounds(*args, **kargs):
            offsprings = func(*args, **kargs)
            for child in offsprings:
                for i in xrange(len(child)):
                    if child[i] > max:
                        child[i] = max
                    elif child[i] < min:
                        child[i] = min
            return offsprings
        return wrapCheckBounds
    return decCheckBounds

toolbox.register("evaluate", benchmarks.kursawe)
toolbox.register("mate", tools.cxBlend, alpha=1.5)
toolbox.register("mutate", tools.mutGaussian, mu=0, sigma=3, indpb=0.3)
toolbox.register("select", tools.selNSGA2)

toolbox.decorate("mate", checkBounds(-5, 5))
toolbox.decorate("mutate", checkBounds(-5, 5)) 

def main():
    random.seed(64)

    MU, LAMBDA = 50, 100
    pop = toolbox.population(n=MU)
    hof = tools.ParetoFront()
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)
    
    algorithms.eaMuPlusLambda(toolbox, pop, mu=MU, lambda_=LAMBDA, 
                              cxpb=0.5, mutpb=0.2, ngen=50, 
                              stats=stats, halloffame=hof)
    
    logging.info("Best individual for measure 1 is %s, %s", 
                 hof[0], hof[0].fitness.values)
    logging.info("Best individual for measure 2 is %s, %s", 
                 hof[-1], hof[-1].fitness.values)

    return pop, stats, hof

if __name__ == "__main__":
    main()



#    This file is part of EAP.
#
#    EAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    EAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with EAP. If not, see <http://www.gnu.org/licenses/>.

import random
import array

from deap import base
from deap import benchmarks
from deap import creator
from deap import tools

# Problem dimension
NDIM = 10

creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
creator.create("Individual", array.array, typecode='d', fitness=creator.FitnessMin)

toolbox = base.Toolbox()
toolbox.register("attr_float", random.uniform, -3, 3)
toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_float, NDIM)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)
toolbox.register("select", tools.selRandom, k=3)
toolbox.register("evaluate", benchmarks.sphere)

def main():
    # Differential evolution parameters
    CR = 0.25
    F = 1  
    MU = 300
    NGEN = 200    
    
    pop = toolbox.population(n=MU);
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)
    
    # Evaluate the individuals
    fitnesses = toolbox.map(toolbox.evaluate, pop)
    for ind, fit in zip(pop, fitnesses):
        ind.fitness.values = fit
    
    for g in range(NGEN):
        for k, agent in enumerate(pop):
            a,b,c = toolbox.select(pop)
            y = toolbox.clone(agent)
            index = random.randrange(NDIM)
            for i, value in enumerate(agent):
                if i == index or random.random() < CR:
                    y[i] = a[i] + F*(b[i]-c[i])
            y.fitness.values = toolbox.evaluate(y)
            if y.fitness > agent.fitness:
                pop[k] = y
        hof.update(pop)
        stats.update(pop)
        
        print "-- Generation %i --" % g
        print stats
            
    print "Best individual is ", hof[0], hof[0].fitness.values[0]
            
if __name__ == "__main__":
    main()



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import array
import random

from deap import algorithms
from deap import base
from deap import creator
from deap import tools

creator.create("FitnessMax", base.Fitness, weights=(1.0,))
creator.create("Individual", array.array, typecode='b', fitness=creator.FitnessMax)

toolbox = base.Toolbox()

# Attribute generator
toolbox.register("attr_bool", random.randint, 0, 1)

# Structure initializers
toolbox.register("individual", tools.initRepeat, creator.Individual, 
    toolbox.attr_bool, 100)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

def evalOneMax(individual):
    return sum(individual),

toolbox.register("evaluate", evalOneMax)
toolbox.register("mate", tools.cxTwoPoints)
toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
toolbox.register("select", tools.selTournament, tournsize=3)
toolbox.register("migrate", tools.migRing, k=5, selection=tools.selBest,
    replacement=tools.selRandom)

def main():
    random.seed(64)

    NBR_DEMES = 3
    MU = 300
    NGEN = 40
    CXPB = 0.5
    MUTPB = 0.2
    MIG_RATE = 5    
    
    demes = [toolbox.population(n=MU) for _ in xrange(NBR_DEMES)]
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values, 4)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)    
    
    for idx, deme in enumerate(demes):
        for ind in deme:
            ind.fitness.values = toolbox.evaluate(ind)
        stats.update(deme, idx)
        hof.update(deme)
    stats.update(demes[0]+demes[1]+demes[2], 3)
    
    gen = 1
    while gen <= NGEN and stats.Max(3)[0] < 100.0:
        print "-- Generation %i --" % gen        
        for idx, deme in enumerate(demes):
            print "-- Deme %i --" % (idx+1)  
            deme[:] = [toolbox.clone(ind) for ind in toolbox.select(deme, len(deme))]
            algorithms.varSimple(toolbox, deme, cxpb=CXPB, mutpb=MUTPB)
            
            for ind in deme:
                ind.fitness.values = toolbox.evaluate(ind)
            
            stats.update(deme, idx)
            hof.update(deme)
            print stats[idx]
        if gen % MIG_RATE == 0:
            toolbox.migrate(demes)
        stats.update(demes[0]+demes[1]+demes[2], 3)
        print "-- Population --"
        print stats[3]
        gen += 1
    
    return demes, stats, hof

if __name__ == "__main__":
    main()



#    This file is part of EAP.
#
#    EAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    EAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with EAP. If not, see <http://www.gnu.org/licenses/>.

import sys
import operator
import math
import logging
import random

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap import gp

# Define new functions
def safeDiv(left, right):
    try:
        return left / right
    except ZeroDivisionError:
        return 0

pset = gp.PrimitiveSet("MAIN", 1)
pset.addPrimitive(operator.add, 2)
pset.addPrimitive(operator.sub, 2)
pset.addPrimitive(operator.mul, 2)
pset.addPrimitive(safeDiv, 2)
pset.addPrimitive(operator.neg, 1)
pset.addPrimitive(math.cos, 1)
pset.addPrimitive(math.sin, 1)
pset.addEphemeralConstant(lambda: random.randint(-1,1))
pset.renameArguments({"ARG0" : "x"})

creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMin, pset=pset)

toolbox = base.Toolbox()
toolbox.register("expr", gp.genRamped, pset=pset, min_=1, max_=2)
toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)
toolbox.register("lambdify", gp.lambdify, pset=pset)

def evalSymbReg(individual):
    # Transform the tree expression in a callable function
    func = toolbox.lambdify(expr=individual)
    # Evaluate the sum of squared difference between the expression
    # and the real function : x**4 + x**3 + x**2 + x
    values = (x/10. for x in xrange(-10,10))
    diff_func = lambda x: (func(x)-(x**4 + x**3 + x**2 + x))**2
    diff = sum(map(diff_func, values))
    return diff,

toolbox.register("evaluate", evalSymbReg)
toolbox.register("select", tools.selTournament, tournsize=3)
toolbox.register("mate", gp.cxUniformOnePoint)
toolbox.register("expr_mut", gp.genFull, min_=0, max_=2)
toolbox.register('mutate', gp.mutUniform, expr=toolbox.expr_mut)

def main():
    random.seed(318)
    
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

    pop = toolbox.population(n=300)
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("Avg", tools.mean)
    stats.register("Std", tools.std)
    stats.register("Min", min)
    stats.register("Max", max)
    
    algorithms.eaSimple(toolbox, pop, 0.5, 0.1, 40, stats, halloffame=hof)
    logging.info("Best individual is %s, %s", gp.evaluate(hof[0]), hof[0].fitness)
    
    return pop, stats, hof

if __name__ == "__main__":
    main()



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

from multiprocessing.connection import Client,Listener
import Queue
import time
import threading
import sys
import os
import socket
import subprocess
from deap.dtm.abstractCommManager import AbstractCommThread

DTM_TCP_MIN_LATENCY = 0.005
DTM_TCP_MAX_LATENCY = 0.01
DTM_CONCURRENT_RECV_LIMIT = 1000
DTM_CONCURRENT_SEND_LIMIT = 1000

BASE_COMM_PORT = 10011

class CommThread(AbstractCommThread):
    
    def __init__(self, recvQ, sendQ, mainThreadEvent, exitEvent, commReadyEvent, randomGenerator, cmdlineArgs):
        AbstractCommThread.__init__(self, recvQ, sendQ, mainThreadEvent, exitEvent, commReadyEvent, randomGenerator, cmdlineArgs)

        self.props = dict(dtmTCPGlobalLaunch = True, 
                        dtmTCPLocalLaunch = False, 
                        dtmTCPLineNo = -1, 
                        dtmTCPhosts = '', 
                        dtmTCPstartPort = BASE_COMM_PORT, 
                        dtmTCPnPortsByHost = -1,
                        dtmTCPnbrWorkers = -1,
                        dtmTCPmaxWorkersByHost = -1,
                        dtmTCPisRootWorker = False)
                        
        self.userArgs = ""
        indexArg = 1
        while indexArg < len(cmdlineArgs):
            keyDataF = cmdlineArgs[indexArg][2:].split('=')
            if keyDataF[0] in self.props.keys():
                self.props[keyDataF[0]] = keyDataF[1]
            else:
                self.userArgs += cmdlineArgs[indexArg] + " "
            indexArg += 1
        if self.props['dtmTCPhosts'] == '':
            cmdlineArgs = cmdlineArgs[1].split(' ')
            self.userArgs = ""
            indexArg = 1
            while indexArg < len(cmdlineArgs):
                keyDataF = cmdlineArgs[indexArg][2:].split('=')
                if keyDataF[0] in self.props.keys():
                    self.props[keyDataF[0]] = keyDataF[1]
                else:
                    self.userArgs += cmdlineArgs[indexArg] + " "
                indexArg += 1
        
    @property
    def poolSize(self):
        return len(self.tabConn)+1      # +1 because tabConn does not include ourselves

    @property
    def workerId(self):
        return (self.workerHost, self.workerHostId)

    @property
    def isRootWorker(self):
        return bool(int(self.props['dtmTCPisRootWorker']))
        
    @property
    def isLaunchProcess(self):
        return bool(int(self.props['dtmTCPGlobalLaunch'])) or bool(int(self.props['dtmTCPLocalLaunch']))

    def iterOverIDs(self):
        return self.tabConn.keys() + [self.workerId]
    
    def setTraceModeOn(self, xmlLog):
        # Profiling is not currently enabled with TCP
        return
    
    def run(self):
        # Looking if we are in "launch mode" or "execute mode"
        if int(self.props['dtmTCPGlobalLaunch']):
            # Global launch. Create a SSH connection with each node, and
            # execute a local launch on them. Then wait for termination
            sshWait = []
            lineTab = []
            maxWorkersByHost = 1
            totalWorkers = 0
            with open(self.props['dtmTCPhosts']) as hostFile:
                for line in hostFile:
                    if line[0] == '#' or line[0] == '\n' or line[0] == '\r':
                        continue
                    lineTab.append(line.replace("\n", "").replace("\r", "").split(" "))
                    totalWorkers += int(lineTab[-1][1].split("=")[1])
                    if int(lineTab[-1][1].split("=")[1]) > maxWorkersByHost:
                        maxWorkersByHost = int(lineTab[-1][1].split("=")[1])
                    
                for lineCount,hostH in enumerate(lineTab):
                    if hostH[0] == 'localhost' or hostH[0] == '127.0.0.1' or hostH[0] == socket.gethostname():
                        # Local launch
                        if len(lineTab) > 1:
                            sys.stderr.write("Warning : 'localhost' used as a global hostname!\nDTM will probably hang.\nReplace localhost by the fully qualified name or run only on this host.")
                            sys.stderr.flush()
                        sshWait.append(subprocess.Popen(["python"+str(sys.version_info[0])+"."+str(sys.version_info[1]), self.cmdArgs[0], 
                                " --dtmTCPGlobalLaunch=0 --dtmTCPLocalLaunch=1 --dtmTCPLineNo="+str(lineCount) +\
                                " --dtmTCPhosts=" + str(self.props['dtmTCPhosts']) + " --dtmTCPnPortsByHost=" + str(maxWorkersByHost*(totalWorkers-1)) +\
                                " --dtmTCPstartPort="+ str(self.props['dtmTCPstartPort']) + " --dtmTCPnbrWorkers="+str(hostH[1].split("=")[1])+\
                                " --dtmTCPmaxWorkersByHost="+ str(maxWorkersByHost)+" "+ self.userArgs], stdout=sys.stdout, stderr=sys.stderr))
                    else:
                        # Remote launch
                        sshWait.append(subprocess.Popen(["/usr/bin/ssh", "-x", hostH[0], "cd "+str(os.getcwd()) +";", "python"+\
                                str(sys.version_info[0])+"."+str(sys.version_info[1])+" " + self.cmdArgs[0] +\
                                " --dtmTCPGlobalLaunch=0 --dtmTCPLocalLaunch=1 --dtmTCPLineNo="+str(lineCount) +\
                                " --dtmTCPhosts=" + str(self.props['dtmTCPhosts']) + " --dtmTCPnPortsByHost=" + str(maxWorkersByHost*(totalWorkers-1)) +\
                                " --dtmTCPstartPort="+ str(self.props['dtmTCPstartPort']) + " --dtmTCPnbrWorkers="+str(hostH[1].split("=")[1])+\
                                " --dtmTCPmaxWorkersByHost="+ str(maxWorkersByHost)+" "+ self.userArgs], stdout=sys.stdout, stderr=sys.stderr))
           
            for connJob in sshWait:
                connJob.wait()
            self.commReadyEvent.set()
            return
        elif int(self.props['dtmTCPLocalLaunch']):
            # Local launch. Makes use of subprocess.popen to launch DTM workers
            # After that, wait until termination
            dtmLocalWorkersList = []
            for i in range(int(self.props['dtmTCPnbrWorkers'])):
                
                if i == 0 and int(self.props['dtmTCPLineNo']) == 0:
                    dtmLocalWorkersList.append(subprocess.Popen(["python"+str(sys.version_info[0])+"."+str(sys.version_info[1]), self.cmdArgs[0],
                        " --dtmTCPGlobalLaunch=0 --dtmTCPLocalLaunch=0 --dtmTCPLineNo="+str(self.props['dtmTCPLineNo']) +\
                        " --dtmTCPhosts=" + str(self.props['dtmTCPhosts']) + " --dtmTCPnPortsByHost=" + str(self.props['dtmTCPnPortsByHost']) +\
                        " --dtmTCPstartPort="+ str(self.props['dtmTCPstartPort']) + " --dtmTCPnbrWorkers="+str(i)+\
                        " --dtmTCPmaxWorkersByHost="+ str(self.props['dtmTCPmaxWorkersByHost']) +" --dtmTCPisRootWorker=1", self.userArgs]))
                else:
                    dtmLocalWorkersList.append(subprocess.Popen(["python"+str(sys.version_info[0])+"."+str(sys.version_info[1]), self.cmdArgs[0],
                        " --dtmTCPGlobalLaunch=0 --dtmTCPLocalLaunch=0 --dtmTCPLineNo="+str(self.props['dtmTCPLineNo']) +\
                        " --dtmTCPhosts=" + str(self.props['dtmTCPhosts']) + " --dtmTCPnPortsByHost=" + str(self.props['dtmTCPnPortsByHost']) +\
                        " --dtmTCPstartPort="+ str(self.props['dtmTCPstartPort']) + " --dtmTCPnbrWorkers="+str(i)+\
                        " --dtmTCPmaxWorkersByHost="+ str(self.props['dtmTCPmaxWorkersByHost']), self.userArgs]))
                                   
            for dtmW in dtmLocalWorkersList:
                dtmW.wait()
            
            self.commReadyEvent.set()
            return
        
        # Here, we are ready to start DTM
        # We have to open the TCP ports and create the communication sockets
        # This is somewhat tricky, because two workers on the same node
        # must not have the same port number, and we do not want to bind
        # n^2 different ports (which will be the trivial way).
        #
        # Each worker as an ID which is a tuple (host, unique number)
        self.tabConn = {}
        listListeners = {}
        lineTab = []
        self.workerHostId = int(self.props['dtmTCPnbrWorkers'])
        maxWorkersByHost = int(self.props['dtmTCPmaxWorkersByHost'])
        fullyhost = "boumbadaboum"
        with open(self.props['dtmTCPhosts']) as hostFile:
            for hostline in hostFile:
                if hostline[0] == '#' or hostline[0] == '\n' or hostline[0] == '\r':
                    continue
                
                lineTab.append(hostline.replace("\n", "").replace("\r", "").split(" "))
                if len(lineTab)-1 == int(self.props['dtmTCPLineNo']):
                    fullyhost = lineTab[-1][0]
                
                # If our line number is GREATER than the other OR our uniq ID is
                # GREATER than the ID of another worker on our node,
                # we create the Client which will attempt to connect to the
                # listener opened by the other.
            
            for lineC,hostInfos in enumerate(lineTab):
                remoteIp = socket.gethostbyname(hostInfos[0])
                
                if lineC > int(self.props['dtmTCPLineNo']):
                    for i in range(0, int(hostInfos[1].split("=")[1])):
                        try:
                            listListeners[(hostInfos[0], i)] = Listener((fullyhost, int(self.props['dtmTCPstartPort']) +\
                                        lineC*maxWorkersByHost + i + self.workerHostId*maxWorkersByHost*len(lineTab)))
                        except Exception as ex:
                            sys.stderr.write(str(("ERROR IN : ", "Listener", self.props['dtmTCPLineNo'], self.workerHostId, lineC, i, maxWorkersByHost, len(lineTab), " < ", remoteIp,
                                        lineC*maxWorkersByHost + i + self.workerHostId*maxWorkersByHost*len(lineTab))) + "\n\n")
                            raise ex
                elif lineC == int(self.props['dtmTCPLineNo']):
                    self.workerHost = hostInfos[0]
                    for i in range(0, int(hostInfos[1].split("=")[1])):
                        if i == self.workerHostId:
                            continue    # Pas de connexion avec nous-memes
                        
                        if i > self.workerHostId:
                            try:
                                listListeners[(hostInfos[0], i)] = Listener(('127.0.0.1', int(self.props['dtmTCPstartPort']) + lineC*maxWorkersByHost + i + self.workerHostId*maxWorkersByHost*len(lineTab)))
                            except Exception as ex:
                                sys.stderr.write(str(("ERROR IN : ", "Listener", self.props['dtmTCPLineNo'], self.workerHostId, lineC, i, maxWorkersByHost, len(lineTab), " => ", remoteIp,
                                        int(self.props['dtmTCPLineNo'])*maxWorkersByHost + i + self.workerHostId*maxWorkersByHost*len(lineTab))) +"\n\n\n")
                                raise ex
                        else:
                            self.tabConn[(hostInfos[0], i)] = Client(('127.0.0.1', int(self.props['dtmTCPstartPort']) + lineC*maxWorkersByHost + self.workerHostId + i*maxWorkersByHost*len(lineTab)))
                            
                else:
                    # lineC < int(self.props['dtmTCPLineNo'])
                    for i in range(0, int(hostInfos[1].split("=")[1])):
                        self.tabConn[(hostInfos[0], i)] = Client((remoteIp, int(self.props['dtmTCPstartPort']) +\
                                        int(self.props['dtmTCPLineNo'])*maxWorkersByHost + self.workerHostId + i*maxWorkersByHost*len(lineTab)))
        
        
        for le in listListeners.keys():
            conn = listListeners[le].accept()
            self.tabConn[le] = conn

        self.commReadyEvent.set()
        
        working = True
        countRecvNotTransmit = 0
        countRecvTimeInit = time.time()

        while working:
            recvSomething = False
            sendSomething = False

            if self.exitStatus.is_set():    # Exit
                # Warning : the communication thread MUST clear the sendQ
                # BEFORE leaving (the exiting orders must be send)
                working = False

            # Check for incoming messages
            for conn in self.tabConn.values():
                if conn.poll():
                    # We received something
                    try:
                        self.recvQ.put(conn.recv())
                        countRecvNotTransmit += 1
                    except EOFError:
                        working = False
                    recvSomething = True
            
            if countRecvNotTransmit > 50 or (time.time() - countRecvTimeInit > 0.1 and countRecvNotTransmit > 0):
                countRecvNotTransmit = 0
                countRecvTimeInit = time.time()
                self.wakeUpMainThread.set()

            while True:
                # Send all our messages
                try:
                    sendMsg = self.sendQ.get_nowait()
                    self.tabConn[sendMsg.receiverWid].send(sendMsg)
                    sendSomething = True
                except Queue.Empty:
                    break

            if not recvSomething:
                time.sleep(self.random.uniform(DTM_TCP_MIN_LATENCY, DTM_TCP_MAX_LATENCY))



from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from matplotlib.colors import LogNorm
import matplotlib.pyplot as plt
import sys

try:
    import numpy as np
except:
    exit()


sys.path.append("../..")
from deap import benchmarks


def untuple(sol):
    return benchmarks.himmelblau(sol)[0]

fig = plt.figure()
ax = Axes3D(fig, azim = -29, elev = 49)
X = np.arange(-6, 6, 0.1)
Y = np.arange(-6, 6, 0.1)
X, Y = np.meshgrid(X, Y)
Z = np.array(map(untuple, zip(X,Y)))

ax.plot_surface(X, Y, Z, rstride = 1, cstride = 1, norm = LogNorm(), cmap = cm.jet, linewidth=0.2, )
 
plt.xlabel("x")
plt.ylabel("y")

plt.show()


try:
    import Queue
except ImportError:
    import queue as Queue
import time
import threading
try:
    import cPickle
except ImportError:
    import pickle as cPickle
import array
import copy
import logging

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        # Python 2.5
        import xml.etree.ElementTree as etree

from deap.dtm.dtmTypes import *
from deap.dtm.abstractCommManager import AbstractCommThread

_logger = logging.getLogger("dtm.communication")

DTM_MPI_MIN_LATENCY = 0.005
DTM_MPI_MAX_LATENCY = 0.01
DTM_CONCURRENT_RECV_LIMIT = 1000
DTM_CONCURRENT_SEND_LIMIT = 1000

class CommThread(AbstractCommThread):

    def __init__(self, recvQ, sendQ, mainThreadEvent, exitEvent, commReadyEvent, randomGenerator, cmdlineArgs):        
        AbstractCommThread.__init__(self, recvQ, sendQ, mainThreadEvent, exitEvent, commReadyEvent, randomGenerator, cmdlineArgs)
    
    @property
    def poolSize(self):
        return self.pSize

    @property
    def workerId(self):
        return self.currentId

    @property
    def isRootWorker(self):
        return self.currentId == 0
    
    @property
    def isLaunchProcess(self):
        return False
    
    def setTraceModeOn(self, xmlLogger):
        self.traceMode = True
        self.traceTo = xmlLogger

    def iterOverIDs(self):
        return range(self.pSize)

    def run(self):
        from mpi4py import MPI
        
        def mpiSend(msg, dest):
            # Pickle and send over MPI
            arrayBuf = array.array('b')
            arrayBuf.fromstring(cPickle.dumps(msg, cPickle.HIGHEST_PROTOCOL))
            
            b = MPI.COMM_WORLD.Isend([arrayBuf, MPI.CHAR], dest=dest, tag=self.msgSendTag)
            if self.traceMode:
                etree.SubElement(self.traceTo, "msg", {"direc" : "out", "type" : str(msg.msgType), "otherWorker" : str(dest), "msgtag" : str(self.msgSendTag), "time" : repr(time.time())})
            
            self.msgSendTag += 1
            return b, arrayBuf
        
        assert MPI.Is_initialized(), "Error in MPI Init!"
        
        self.pSize = MPI.COMM_WORLD.Get_size()
        self.currentId = MPI.COMM_WORLD.Get_rank()
        
        self.commReadyEvent.set()   # Notify the main thread that we are ready
        
        if self.currentId == 0 and MPI.Query_thread() > 0:
            # Warn only once
            _logger.warning("MPI was initialized with a thread level of %i, which is higher than MPI_THREAD_SINGLE."
            " The current MPI implementations do not always handle well the MPI_THREAD_MULTIPLE or MPI_THREAD_SERIALIZED modes."
            " As DTM was designed to work with the base, safe mode (MPI_THREAD_SINGLE), it is strongly suggested to change"
            " the 'thread_level' variable or your mpi4py settings in 'site-packages/mpi4py/rc.py', unless you have strong"
            " motivations to keep that setting. This may bring both stability and performance improvements.", MPI.Query_thread())
        
        lRecvWaiting = []
        lSendWaiting = []
        countSend = 0
        countRecv = 0
        lMessageStatus = MPI.Status()
        working = True
        
        countRecvNotTransmit = 0
        countRecvTimeInit = time.time()

        while working:
            recvSomething = False
            sendSomething = False

            if self.exitStatus.is_set():    # Exiting
                # Warning : the communication thread MUST clear the sendQ
                # BEFORE leaving (the exiting orders must be send)
                working = False

            while len(lRecvWaiting) < DTM_CONCURRENT_RECV_LIMIT and MPI.COMM_WORLD.Iprobe(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=lMessageStatus):
                # We received something
                lBuf = array.array('b', (0,))
                lBuf = lBuf * lMessageStatus.Get_elements(MPI.CHAR)
                
                lRecvWaiting.append((lBuf, MPI.COMM_WORLD.Irecv([lBuf, MPI.CHAR], source=lMessageStatus.Get_source(), tag=lMessageStatus.Get_tag()), lMessageStatus.Get_tag()))

                lMessageStatus = MPI.Status()
                recvSomething = True

            
            for i, reqTuple in enumerate(lRecvWaiting):
                if reqTuple[1].Test():
                    countRecv += 1
                    dataS = cPickle.loads(reqTuple[0].tostring())
                    if self.traceMode:
                        etree.SubElement(self.traceTo, "msg", {"direc" : "in", "type" : str(dataS.msgType), "otherWorker" : str(dataS.senderWid), "msgtag" : str(reqTuple[2]), "time" : repr(time.time())})
                    self.recvQ.put(dataS)
                    lRecvWaiting[i] = None
                    recvSomething = True
                    # Wake up the main thread if there's a sufficient number
                    # of pending receives
                    countRecvNotTransmit += 1
                    
                    
            if countRecvNotTransmit > 50 or (time.time() - countRecvTimeInit > 0.1 and countRecvNotTransmit > 0):
                countRecvNotTransmit = 0
                countRecvTimeInit = time.time()
                self.wakeUpMainThread.set()        
                    
            lRecvWaiting = filter(lambda d: not d is None, lRecvWaiting)
            if not isinstance(lRecvWaiting, list):
                lRecvWaiting = list(lRecvWaiting)

            while len(lSendWaiting) < DTM_CONCURRENT_SEND_LIMIT:
                # Send all pending sends, under the limit of
                # DTM_CONCURRENT_SEND_LIMIT
                try:
                    sendMsg = self.sendQ.get_nowait()
                    countSend += 1
                    sendMsg.sendTime = time.time()
                    commA, buf1 = mpiSend(sendMsg, sendMsg.receiverWid)
                    lSendWaiting.append((commA, buf1))
                    sendSomething = True
                except Queue.Empty:
                    break
            
            lSendWaiting = filter(lambda d: not d[0].Test(), lSendWaiting)
            if not isinstance(lSendWaiting, list):  # Python 3
                lSendWaiting = list(lSendWaiting)
            
            if not recvSomething:
                time.sleep(self.random.uniform(DTM_MPI_MIN_LATENCY, DTM_MPI_MAX_LATENCY))
                
        while len(lSendWaiting) > 0:
            # Send the lasts messages before shutdown
            lSendWaiting = filter(lambda d: not d[0].Test(), lSendWaiting)
            if not isinstance(lSendWaiting, list):  # Python 3
                lSendWaiting = list(lSendWaiting)
            time.sleep(self.random.uniform(DTM_MPI_MIN_LATENCY, DTM_MPI_MAX_LATENCY))
            
        del lSendWaiting
        del lRecvWaiting
        